from unittest import TestCase

from airflow_mcd.operators.dbt import DbtConfig


class DbtCliConfigTests(TestCase):
    def test_build_command_with_default_config(self):
        # given
        config = DbtConfig()

        # when
        command = config.build_command("run")

        # then
        self.assertEqual([
            "dbt",
            "run",
        ], command)

    def test_build_command_with_non_default_config(self):
        # given
        config = DbtConfig(
            dbt_bin="/usr/bin/dbt",
            profiles_dir="/usr/local/dbt/.profiles",
            target_path="/tmp/dbt/target",
            target="prod",
            vars={"foo": "bar"},
            models="model",
            exclude="excluded_model",
            select="selected_model",
            warn_error=True,
            full_refresh=True,
            data=True,
            schema=True,
        )

        # when
        command = config.build_command("docs generate")

        # then
        self.assertEqual([
            "/usr/bin/dbt",
            "--warn-error",
            "docs",
            "generate",
            "--profiles-dir",
            "/usr/local/dbt/.profiles",
            "--target",
            "prod",
            "--vars",
            "{\"foo\": \"bar\"}",
            "--models",
            "model",
            "--exclude",
            "excluded_model",
            "--select",
            "selected_model",
            "--full-refresh",
            "--data",
            "--schema",
        ], command)



import importlib
import json
import logging
import os
import re
import subprocess
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

from airflow.exceptions import AirflowException
from airflow.models import BaseOperator
from airflow.utils.context import Context

from airflow_mcd.hooks import SessionHook

from pycarlo.core import Client
from pycarlo.features.dbt import DbtImporter

logger = logging.getLogger(__name__)


class DbtConfig:
    """
    Configuration for the execution of a dbt operator.

    :param mc_conn_id: Identifier of Monte Carlo connection configured in Airflow
    :param mc_include_logs: If `True`, write a copy of the logs generated by the dbt command to a local file and include with artifacts sent to Monte Carlo
    :param dir: The directory to run the CLI in
    :param dbt_bin: The `dbt` CLI. Defaults to `dbt`, so assumes it's on your `PATH`
    :param env: Optional environment variables to add to dbt command execution
    :param profiles_dir: If set, passed as the `--profiles-dir` argument to the `dbt` command
    :param target_path: If set, overrides expected location of dbt artifacts (default is `./target`)
    :param target: If set, passed as the `--target` argument to the `dbt` command
    :param vars: If set, passed as the `--vars` argument to the `dbt` command
    :param models: If set, passed as the `--models` argument to the `dbt` command
    :param exclude: If set, passed as the `--exclude` argument to the `dbt` command
    :param select: If set, passed as the `--select` argument to the `dbt` command
    :param verbose: The operator will log verbosely to the Airflow logs
    :param warn_error: If `True`, treat warnings as errors.
    :param full_refresh: If `True`, will fully-refresh incremental models.
    :param data: If `True`, includes `--data` in the `dbt` command
    :param schema: If `True`, includes `--schema` in the `dbt` command
    """
    def __init__(
            self,
            mc_conn_id: Optional[str] = None,
            mc_include_logs: bool = True,
            dir: str = ".",
            dbt_bin: str = "dbt",
            env: Optional[Dict] = None,
            profiles_dir: Optional[str] = None,
            target_path: Optional[str] = None,
            target: Optional[str] = None,
            vars: Optional[Dict] = None,
            models: Optional[str] = None,
            exclude: Optional[str] = None,
            select: Optional[str] = None,
            verbose: bool = True,
            warn_error: bool = False,
            full_refresh: bool = False,
            data: bool = False,
            schema: bool = False,
    ):
        self.mc_conn_id = mc_conn_id
        self.mc_include_logs = mc_include_logs
        self.dir = dir
        self.dbt_bin = dbt_bin
        self.env = {} if env is None else env
        self.profiles_dir = profiles_dir
        self.target_path = target_path
        self.target = target
        self.vars = vars
        self.models = models
        self.exclude = exclude
        self.select = select
        self.verbose = verbose
        self.warn_error = warn_error
        self.full_refresh = full_refresh
        self.data = data
        self.schema = schema

    def build_command(self, command: str) -> List[str]:
        """
        Build dbt command

        :param command: dbt command, e.g. "run" or "docs generate"
        """
        dbt_cmd = [self.dbt_bin] + command.split()

        if self.profiles_dir is not None:
            dbt_cmd.extend(['--profiles-dir', self.profiles_dir])

        if self.target is not None:
            dbt_cmd.extend(['--target', self.target])

        if self.vars is not None:
            dbt_cmd.extend(['--vars', json.dumps(self.vars)])

        if self.models is not None:
            dbt_cmd.extend(['--models', self.models])

        if self.exclude is not None:
            dbt_cmd.extend(['--exclude', self.exclude])

        if self.select is not None:
            dbt_cmd.extend(['--select', self.select])

        if self.full_refresh:
            dbt_cmd.extend(['--full-refresh'])

        if self.data:
            dbt_cmd.extend(['--data'])

        if self.schema:
            dbt_cmd.extend(['--schema'])

        if self.warn_error:
            dbt_cmd.insert(1, '--warn-error')

        return dbt_cmd


class DefaultConfigProvider(ABC):
    """
    Concrete implementations of this class will provide a default dbt CLI configuration for each dbt operator.

    The fully qualified name of the concrete class should be set in a `AIRFLOW_MCD_DBT_CONFIG_PROVIDER` environment
    variable in the Airflow runtime environment.

    AWS MWAA users should add an `mc.airflow_mcd_dbt_config_provider` custom configuration property to the MWAA environment.
    """
    ENV_VAR_NAME = "AIRFLOW_MCD_DBT_CONFIG_PROVIDER"

    @abstractmethod
    def config(self) -> DbtConfig:
        """
        Concrete implementation should return a default dbt CLI configuration.
        """
        pass

    @classmethod
    def get(cls) -> DbtConfig:
        """
        Get the default dbt CLI configuration.
        (from fully qualified class name defined in `AIRFLOW_MCD_DEFAULT_CONFIG_PROVIDER`).
        """
        provider_fqn = os.getenv(cls.ENV_VAR_NAME) or \
            os.getenv(f"AIRFLOW__MC__{cls.ENV_VAR_NAME}")  # for AWS MWAA

        if provider_fqn:
            module_and_class_name = provider_fqn.rsplit(".", 1)
            provider_module = importlib.import_module(module_and_class_name[0])
            provider_class = getattr(provider_module, module_and_class_name[1])
            provider = provider_class()
            if isinstance(provider, DefaultConfigProvider):
                logger.info(f"Loading default dbt operator configuration from: {provider_fqn}")
                return provider.config()
            else:
                raise Exception(f"dbt configuration provider defined in {cls.ENV_VAR_NAME} is not an instance of {cls.__name__}")

        return DbtConfig()


class DbtOperator(BaseOperator):
    """
    Airflow operator that executes a dbt command and sends artifacts to Monte Carlo.

    :param command: dbt command to execute, e.g. "run", "test", or "build"
    :param project_name: dbt project name to send to Monte Carlo
    :param job_name: dbt job name to send to Monte Carlo
    :param mc_conn_id: Identifier of Monte Carlo connection configured in Airflow
    :param mc_resource_id: Identifier of Monte Carlo lake/warehouse to use to match dbt results to tables (required if you are monitoring multiple lake/warehouse resources in Monte Carlo)
    :param mc_enabled: If `True`, send dbt artifacts to Monte Carlo after dbt command has completed
    :param mc_include_logs: If `True`, write a copy of the logs generated by the dbt command to a local file and include with artifacts sent to Monte Carlo
    :param mc_success_required: If `True`, Airflow task will fail if dbt artifacts cannot be sent to Monte Carlo
    :param dbt_success_required: If `True`, Airflow task will fail if dbt command returns a non-zero status code
    :param dir: The directory to run the CLI in
    :param dbt_bin: The `dbt` CLI. Defaults to `dbt`, so assumes it's on your `PATH`
    :param env: Optional environment variables to add to dbt command execution
    :param profiles_dir: If set, passed as the `--profiles-dir` argument to the `dbt` command
    :param target_path: If set, overrides expected location of dbt artifacts (default is `./target`)
    :param target: If set, passed as the `--target` argument to the `dbt` command
    :param vars: If set, passed as the `--vars` argument to the `dbt` command
    :param models: If set, passed as the `--models` argument to the `dbt` command
    :param exclude: If set, passed as the `--exclude` argument to the `dbt` command
    :param select: If set, passed as the `--select` argument to the `dbt` command
    :param verbose: The operator will log verbosely to the Airflow logs
    :param warn_error: If `True`, treat warnings as errors.
    :param full_refresh: If `True`, will fully-refresh incremental models.
    :param data: If `True`, includes `--data` in the `dbt` command
    :param schema: If `True`, includes `--schema` in the `dbt` command
    """

    # color of nodes in the Airflow UI (blue from MC logo)
    ui_color = "#248BF0"

    # allow use of Jinja templates in these fields
    template_fields = ['vars']

    # dbt spits out some escape characters in the logs for some reason, this regex is used to remove them
    ESCAPE_REGEX = re.compile(r'\x1b')

    def __init__(self,
                 command: str,
                 project_name: str,
                 job_name: str,
                 mc_conn_id: Optional[str] = None,
                 mc_resource_id: Optional[str] = None,
                 mc_enabled: bool = True,
                 mc_include_logs: bool = False,
                 mc_success_required: bool = False,
                 dbt_success_required: bool = True,
                 dir: str = None,
                 dbt_bin: str = None,
                 env: Optional[Dict] = None,
                 profiles_dir: Optional[str] = None,
                 target_path: Optional[str] = None,
                 target: Optional[str] = None,
                 vars: Optional[Dict] = None,
                 models: Optional[str] = None,
                 exclude: Optional[str] = None,
                 select: Optional[str] = None,
                 verbose: bool = True,
                 warn_error: bool = False,
                 full_refresh: bool = False,
                 data: bool = False,
                 schema: bool = False,
                 **kwargs):

        super(DbtOperator, self).__init__(**kwargs)
        self.command = command
        self.project_name = project_name
        self.job_name = job_name
        self.mc_conn_id = mc_conn_id
        self.mc_resource_id = mc_resource_id
        self.mc_enabled = mc_enabled
        self.mc_include_logs = mc_include_logs
        self.mc_success_required = mc_success_required
        self.dbt_success_required = dbt_success_required
        self.dir = dir
        self.dbt_bin = dbt_bin
        self.env = env
        self.profiles_dir = profiles_dir
        self.target_path = target_path
        self.target = target
        self.vars = vars
        self.models = models
        self.exclude = exclude
        self.select = select
        self.verbose = verbose
        self.warn_error = warn_error
        self.full_refresh = full_refresh
        self.data = data
        self.schema = schema

    def execute(self, context: Context) -> Any:
        config = self._merge_config()
        log_file_path = f"{config.dir}/dbt.log" if config.mc_include_logs else None
        return_code = self._execute_dbt(config, log_file_path)
        self._send_artifacts_to_monte_carlo(config, log_file_path)
        if return_code > 0 and self.dbt_success_required:
            raise AirflowException("dbt command failed")

    def _merge_config(self) -> DbtConfig:
        # get default config to be applied to all operators
        default_config = DefaultConfigProvider.get()

        # merge operator-provided config with default
        config = DbtConfig(
            mc_conn_id=self.mc_conn_id or default_config.mc_conn_id,
            mc_include_logs=self.mc_include_logs or default_config.mc_include_logs,
            dir=self.dir or default_config.dir,
            dbt_bin=self.dbt_bin or default_config.dbt_bin,
            env=default_config.env,
            profiles_dir=self.profiles_dir or default_config.profiles_dir,
            target_path=self.target_path or default_config.target_path,
            target=self.target or default_config.target,
            vars=self.vars or default_config.vars,
            models=self.models or default_config.models,
            exclude=self.exclude or default_config.exclude,
            select=self.select or default_config.select,
            verbose=self.verbose or default_config.verbose,
            warn_error=self.warn_error or default_config.warn_error,
            full_refresh=self.full_refresh or default_config.full_refresh,
            data=self.data or default_config.data,
            schema=self.schema or default_config.schema,
        )
        if self.env:
            config.env.update(self.env)

        return config

    def _execute_dbt(self, config: DbtConfig, log_file_path: Optional[str]) -> int:
        # build dbt command
        cmd = config.build_command(self.command)
        if config.verbose:
            self.log.info(" ".join(cmd))

        # gather environment variables
        env = os.environ.copy()
        if config.env:
            env.update(config.env)

        # submit dbt command
        process = subprocess.Popen(
            cmd,
            env=env,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            cwd=config.dir,
            close_fds=True,
        )

        # logging, both to Airflow logger and a file (to be sent to Monte Carlo)
        log_file = open(log_file_path, "w") if log_file_path else None
        for line in iter(process.stdout.readline, b""):
            line = self.ESCAPE_REGEX.sub('', line.decode("utf-8"))
            self.log.info(line.rstrip())
            if log_file:
                log_file.write(line)

        # wait for command to finish
        process.wait()
        self.log.info(f"Command exited with return code: {process.returncode}")
        if log_file:
            log_file.close()

        return process.returncode

    def _send_artifacts_to_monte_carlo(self, config: DbtConfig, log_file_path: Optional[str]):
        if self.mc_enabled:
            target_path = config.target_path or f"{config.dir}/target"
            self.log.info(f"Sending dbt artifacts in '{target_path}' to Monte Carlo")
            try:
                DbtImporter(
                    mc_client=Client(session=SessionHook(mcd_session_conn_id=config.mc_conn_id).get_conn())
                ).import_run(
                    manifest_path=f"{target_path}/manifest.json",
                    run_results_path=f"{target_path}/run_results.json",
                    logs_path=log_file_path,
                    project_name=self.project_name,
                    job_name=self.job_name,
                    resource_id=self.mc_resource_id,
                )
                self.log.info("Successfully sent dbt artifacts to Monte Carlo")
            except Exception as ex:
                self.log.warning(f"Failed to send dbt artifacts to Monte Carlo: {ex}")
                if self.mc_success_required:
                    raise ex


class DbtBuildOperator(DbtOperator):
    def __init__(self, **kwargs):
        super(DbtBuildOperator, self).__init__(command="build", **kwargs)


class DbtRunOperator(DbtOperator):
    def __init__(self, **kwargs):
        super(DbtRunOperator, self).__init__(command="run", **kwargs)


class DbtSeedOperator(DbtOperator):
    def __init__(self, **kwargs):
        super(DbtSeedOperator, self).__init__(command="seed", **kwargs)


class DbtSnapshotOperator(DbtOperator):
    def __init__(self, **kwargs):
        super(DbtSnapshotOperator, self).__init__(command="snapshot", **kwargs)


class DbtTestOperator(DbtOperator):
    def __init__(self, **kwargs):
        super(DbtTestOperator, self).__init__(command="test", **kwargs)



from typing import Tuple

from pycarlo.core import Client
from requests import HTTPError

from airflow_mcd.hooks import SessionHook
from airflow_mcd.hooks.constants import MCD_GATEWAY_CONNECTION_TYPE

_INTEGRATION_GATEWAY_PATH = "/airflow/callbacks"


class GatewaySessionHook(SessionHook):
    conn_name_attr = "mcd_session_conn_id"

    default_conn_name = "mcd_gateway_default_session"

    conn_type = MCD_GATEWAY_CONNECTION_TYPE

    hook_name = "Monte Carlo Data Gateway"

    def test_connection(self) -> Tuple[bool, str]:
        """
        Test MCD Gateway connection.

        :return: tuple of boolean and string for whether the test was successful and a related message
        """
        try:
            mc_client = Client(self.get_conn())
            body = {
                "airflow_operation": "ping",
                "airflow_payload": {},
            }
            json_response = mc_client.make_request(path=_INTEGRATION_GATEWAY_PATH, body=body)
            success = json_response.get("success", False)
            resource_ids = json_response.get("resource_ids", [])

            if not success:
                return False, f"Failed to test MCD connection, returned success=false"

            suffix = f" Using Airflow connection in MCD with UUID={resource_ids[0]}" if resource_ids else ""
            return True, f"Connection to MCD Gateway successfully tested.{suffix}"
        except HTTPError as exc:
            if exc.response.status_code == 401:
                return False, (
                    "Authentication failed, please check you created an Airflow Integration Key following "
                    "the docs at https://docs.getmontecarlo.com/docs/airflow-incidents-dags-and-tasks"
                )
            return False, f"Connection to MCD Gateway failed: {exc}"
        except Exception as exc:
            return False, f"Connection to MCD Gateway failed: {exc}"



import os
from datetime import datetime, timedelta, timezone
from typing import Dict, List, Optional, Tuple
from unittest import TestCase
from unittest.mock import create_autospec, patch, ANY

from airflow.models import DagRun, TaskInstance, SlaMiss, DAG, BaseOperator
from sgqlc.types import Variable

from airflow_mcd.callbacks.client import AirflowEventsClient, AirflowEnv
from airflow_mcd.callbacks.utils import AirflowEventsClientUtils, _EXCEPTION_MSG_LIMIT
from freezegun import freeze_time
from pycarlo.common.utils import truncate_string
from pycarlo.core import Client


# needed to have a successful assert_called_with as Variable doesn't implement __eq__
class EqVariable(Variable):
    def __eq__(self, other):
        return other.name == self.name


class CallbacksTests(TestCase):
    @patch("airflow_mcd.callbacks.client.AirflowEventsClient._upload_result")
    def test_upload_dag_result_success(self,  mock_client_upload_result):
        self._test_upload_dag_result(True, mock_client_upload_result)

    @freeze_time("2023-02-03 10:11:12", tz_offset=0)
    @patch("airflow_mcd.callbacks.client.AirflowEventsClient._upload_result")
    def test_upload_dag_result_success_no_dates(self,  mock_client_upload_result):
        self._test_upload_dag_result(True, mock_client_upload_result, set_dates=False)

    @freeze_time("2023-02-03 10:11:12", tz_offset=0)
    @patch("airflow_mcd.callbacks.client.AirflowEventsClient._upload_result")
    def test_upload_dag_result_success_no_end_date(self,  mock_client_upload_result):
        self._test_upload_dag_result(True, mock_client_upload_result, set_dates=True, set_end_date=False)

    @patch("airflow_mcd.callbacks.client.AirflowEventsClient._upload_result")
    def test_upload_dag_result_failure(self, mock_client_upload_result):
        self._test_upload_dag_result(False, mock_client_upload_result)

    @patch("airflow_mcd.callbacks.client.AirflowEventsClient._upload_result")
    def test_upload_task_result_success(self, mock_client_upload_result):
        self._test_upload_task_result("success", mock_client_upload_result)

    @freeze_time("2023-02-03 10:11:12", tz_offset=0)
    @patch("airflow_mcd.callbacks.client.AirflowEventsClient._upload_result")
    def test_upload_task_result_success_no_dates(self, mock_client_upload_result):
        self._test_upload_task_result("success", mock_client_upload_result, set_dates=False)

    @patch("airflow_mcd.callbacks.client.AirflowEventsClient._upload_result")
    def test_upload_task_result_failure(self, mock_client_upload_result):
        self._test_upload_task_result("failed", mock_client_upload_result)

    @patch("airflow_mcd.callbacks.client.AirflowEventsClient._upload_result")
    def test_upload_task_result_failure_long_message(self, mock_client_upload_result):
        error_message = "error message érror, " * 1024
        self._test_upload_task_result(
            state="failed",
            mock_client_upload_result=mock_client_upload_result,
            error_message=error_message,
        )

    @patch("airflow_mcd.callbacks.client.AirflowEventsClient._upload_result")
    def test_upload_task_result_running(self, mock_client_upload_result):
        self._test_upload_task_result("running", mock_client_upload_result)

    def test_task_instance_with_no_task(self):
        # simulate a task instance with no task in Airflow >= 2.9.0
        task = create_autospec(BaseOperator)
        task.executor = None
        task.task_id = "1"
        task.queue = "q1"
        task.pool_slots = 1
        task.run_as_user = None
        task.executor_config = {}
        task.retries = 0
        task_instance = TaskInstance(task=task)
        task_instance.task = None
        task_instance.end_date = datetime.utcnow()
        self.assertIsNone(AirflowEventsClientUtils._get_next_retry_datetime(task_instance))
        self.assertIsNone(AirflowEventsClientUtils._get_lineage_list(task_instance, "inlets"))


    @patch("airflow_mcd.callbacks.client.AirflowEventsClient._upload_result")
    def test_upload_sla_misses(self, mock_client_upload_result):
        utils = AirflowEventsClientUtils()

        dag = create_autospec(DAG)
        dag.dag_id = "dag_123"
        dag.params = {}

        sla_miss_1 = create_autospec(SlaMiss)
        sla_miss_1.task_id = "task_123"
        sla_miss_1.execution_date = datetime.now(tz=timezone.utc) - timedelta(seconds=10)
        sla_miss_1.timestamp = datetime.now(tz=timezone.utc)

        sla_miss_2 = create_autospec(SlaMiss)
        sla_miss_2.task_id = "task_234"
        sla_miss_2.execution_date = datetime.now(tz=timezone.utc) - timedelta(seconds=10)
        sla_miss_2.timestamp = datetime.now(tz=timezone.utc)

        sla_misses = [
            sla_miss_1,
            sla_miss_2,
        ]
        utils.mcd_post_sla_misses(dag, sla_misses)

        mock_client_upload_result.assert_called()
        mock_client_upload_result.assert_called_with(
            AirflowEventsClient._UPLOAD_AIRFLOW_SLA_MISSES_OPERATION,
            {
                "dag_id": dag.dag_id,
                "env": self._get_graphql_env(),
                "payload": EqVariable("payload"),
            },
            {
                "event_type": "sla_miss",
                "dag_id": dag.dag_id,
                "env": self._get_env(),
                "sla_misses": [
                    {
                        "task_id": m.task_id,
                        "execution_date": m.execution_date.isoformat(),
                        "timestamp": m.timestamp.isoformat(),
                    }
                    for m in sla_misses
                ]
            }
        )

    def test_env_loading(self):
        no_env = AirflowEnv()
        self.assertEqual("airflow", no_env.env_name)
        self.assertIsNone(no_env.env_id)
        self.assertIsNone(no_env.version)
        self.assertIsNone(no_env.base_url)

        # AWS
        with patch.dict(os.environ, {
            "AIRFLOW_ENV_NAME": "aws_env_name",
            "AIRFLOW_ENV_ID": "aws_env_id",
            "AIRFLOW_VERSION": "aws_version",
            "AIRFLOW__WEBSERVER__BASE_URL": "aws_url",
        }):
            env = AirflowEnv()
            self.assertEqual("aws_env_name", env.env_name)
            self.assertEqual("aws_env_id", env.env_id)
            self.assertEqual("aws_version", env.version)
            self.assertEqual("aws_url", env.base_url)

        # GCP Composer
        with patch.dict(os.environ, {
            "COMPOSER_ENVIRONMENT": "gcp_env_name",
            "COMPOSER_GKE_NAME": "gcp_env_id",
            "MAJOR_VERSION": "gcp_version",
            "AIRFLOW__WEBSERVER__BASE_URL": "gcp_url",
        }):
            env = AirflowEnv()
            self.assertEqual("gcp_env_name", env.env_name)
            self.assertEqual("gcp_env_id", env.env_id)
            self.assertEqual("gcp_version", env.version)
            self.assertEqual("gcp_url", env.base_url)

        # Astronomer
        with patch.dict(os.environ, {
            "AIRFLOW__WEBSERVER__INSTANCE_NAME": "astro_env_name",
            "ASTRO_DEPLOYMENT_ID": "astro_env_id",
            "ASTRONOMER_RUNTIME_VERSION": "astro_version",
            "AIRFLOW__WEBSERVER__BASE_URL": "astro_url",
        }):
            env = AirflowEnv()
            self.assertEqual("astro_env_name", env.env_name)
            self.assertEqual("astro_env_id", env.env_id)
            self.assertEqual("astro_version", env.version)
            self.assertEqual("astro_url", env.base_url)

        params_env = AirflowEnv(
            env_name="name",
            env_id="id",
            version="1.0",
            base_url="url"
        )
        self.assertEqual("name", params_env.env_name)
        self.assertEqual("id", params_env.env_id)
        self.assertEqual("1.0", params_env.version)
        self.assertEqual("url", params_env.base_url)

    @patch("airflow_mcd.callbacks.client.AirflowEventsClient._get_existing_connection_id")
    @patch("airflow_mcd.callbacks.client.AirflowEventsClient._get_session")
    @patch("airflow_mcd.callbacks.client.Client")
    def test_upload_igw_dag_result_success(
        self,
        mock_client_factory,
        mock_get_session,
        mock_get_connection_id,
    ):
        mock_client = create_autospec(Client)
        mock_client_factory.return_value = mock_client
        path = "/airflow/callbacks"
        mock_get_connection_id.return_value = "conn_id"

        dag_context, dag, dag_run, task_instances = self._upload_dag_result(True)
        expected_body = {
            "airflow_operation": AirflowEventsClient._UPLOAD_AIRFLOW_DAG_RESULT_OPERATION,
            "airflow_payload": {
                "dag_id": dag.dag_id,
                "run_id": dag_context["run_id"],
                "success": True,
                "reason": dag_context.get("reason"),
                "state": dag_run.state,
                "execution_date": dag_run.execution_date.isoformat(),
                "start_date": dag_run.start_date.isoformat(),
                "end_date": dag_run.end_date.isoformat(),
                "env": self._get_graphql_env(),
                "payload": {
                    "dag_id": dag.dag_id,
                    "env": self._get_env(),
                    "run_id": dag_context["run_id"],
                    "success": True,
                    "tasks": [
                        self._get_dag_task_instance_result(ti, set_dates=True)
                        for ti in task_instances
                    ],
                    "state": dag_run.state,
                    "execution_date": dag_run.execution_date.isoformat(),
                    "start_date": dag_run.start_date.isoformat(),
                    "end_date": dag_run.end_date.isoformat(),
                    "reason": dag_context.get("reason"),
                    "event_type": "dag",
                    "original_dates": ANY,
                },
            },
        }
        mock_client.make_request.assert_called_with(path=path, body=expected_body, timeout_in_seconds=10, should_retry=ANY)
        mock_get_connection_id.assert_called()

    def _test_upload_dag_result(
            self,
            success: bool,
            mock_client_upload_result,
            set_dates: bool = True,
            set_end_date: bool = True,
    ):
        dag_context, dag, dag_run, task_instances = self._upload_dag_result(success, set_dates, set_end_date)
        now_isoformat = datetime.now(tz=timezone.utc).isoformat()
        if not set_end_date:
            mock_client_upload_result.assert_not_called()
            return
        mock_client_upload_result.assert_called()
        mock_client_upload_result.assert_called_with(
            AirflowEventsClient._UPLOAD_AIRFLOW_DAG_RESULT_OPERATION,
            {
                "dag_id": dag.dag_id,
                "run_id": dag_context["run_id"],
                "success": success,
                "reason": dag_context.get("reason"),
                "state": dag_run.state,
                "execution_date": dag_run.execution_date.isoformat() if set_dates else now_isoformat,
                "start_date": dag_run.start_date.isoformat() if set_dates else now_isoformat,
                "end_date": dag_run.end_date.isoformat() if set_dates else now_isoformat,
                "env": self._get_graphql_env(),
                "payload": EqVariable("payload"),
            },
            {
                "dag_id": dag.dag_id,
                "env": self._get_env(),
                "run_id": dag_context["run_id"],
                "success": success,
                "tasks": [
                    self._get_dag_task_instance_result(ti, set_dates=set_dates) for ti in task_instances
                ],
                "state": dag_run.state,
                "execution_date": dag_run.execution_date.isoformat() if set_dates else now_isoformat,
                "start_date": dag_run.start_date.isoformat() if set_dates else now_isoformat,
                "end_date": dag_run.end_date.isoformat() if set_dates else now_isoformat,
                "reason": dag_context.get("reason"),
                "event_type": "dag",
                "original_dates": ANY,
            }
        )

    def _upload_dag_result(
        self, success: bool, set_dates: bool = True, set_end_date: bool = True,
    ) -> Tuple[Dict, DAG, DagRun, List[TaskInstance]]:
        state = "success" if success else "failed"
        dag_context = self._create_dag_context(state, set_dates=set_dates, set_end_date=set_end_date)
        dag: DAG = dag_context["dag"]
        dag_run: DagRun = dag_context["dag_run"]
        task_instances: List[TaskInstance] = dag_run.get_task_instances()

        utils = AirflowEventsClientUtils()
        utils.mcd_post_dag_result(dag_context)
        return dag_context, dag, dag_run, task_instances


    @staticmethod
    def _get_graphql_env() -> Dict:
        return {
            "env_name": "airflow",
        }

    @staticmethod
    def _get_env() -> Dict:
        return {
            "env_name": "airflow",
            "env_id": None,
            "version": None,
            "base_url": None
        }

    def _create_dag_context(
            self,
            state: str,
            set_dates: bool = True,
            set_end_date: bool = True,
    ) -> Dict:
        dag = create_autospec(DAG)
        dag.dag_id = "dag_123"
        dag.params = {}
        dag_run = create_autospec(DagRun)
        task_instances = [
            self._create_task_instance(
                dag_id=dag.dag_id,
                task_id="task_123",
                state=state,
                running=state == "running",
                set_dates=set_dates,
            ),
            self._create_task_instance(
                dag_id=dag.dag_id,
                task_id="task_234",
                state="success",
                set_dates=set_dates,
            ),
        ]
        dag_run.get_task_instances.return_value = task_instances
        dag_run.state = state
        if set_dates:
            dag_run.execution_date = datetime.now(tz=timezone.utc) - timedelta(seconds=10)
            dag_run.start_date = datetime.now(tz=timezone.utc) - timedelta(seconds=9)
        else:
            dag_run.execution_date = None
            dag_run.start_date = None

        if set_end_date:
            dag_run.end_date = datetime.now(tz=timezone.utc)
        else:
            dag_run.end_date = None

        dag_context = {
            "dag": dag,
            "run_id": '123',
            "dag_run": dag_run,
            "reason": "succeeded" if state == "success" else "task failed",
        }
        return dag_context

    def _test_upload_task_result(
        self,
        state: str,
        mock_client_upload_result,
        set_dates: bool = True,
        error_message: Optional[str] = None,
    ):
        dag_context = self._create_dag_context(state, set_dates=set_dates)
        exception_message: str = (error_message or "task failed") if state == "failed" else None
        if state == "failed":
            dag_context["exception"] = Exception(exception_message)
        dag: DAG = dag_context["dag"]
        dag_run: DagRun = dag_context["dag_run"]
        task_instances: List[TaskInstance] = dag_run.get_task_instances()
        task_instance = task_instances[0]
        task_instance.state = state
        dag_context["task_instance"] = task_instance
        utils = AirflowEventsClientUtils()
        utils.mcd_post_task_result(dag_context)

        now_isoformat = datetime.now(tz=timezone.utc).isoformat()
        expected_graphql_payload = {
            "dag_id": dag.dag_id,
            "run_id": dag_context["run_id"],
            "task_id": task_instance.task_id,
            "success": state == "success",
            "env": self._get_graphql_env(),
            "state": state,
            "log_url": f"http://airflow.com/{dag.dag_id}/{task_instance.task_id}/log",
            "execution_date": task_instance.execution_date.isoformat() if set_dates else now_isoformat,
            'start_date': task_instance.start_date.isoformat() if set_dates else now_isoformat,
            'end_date': task_instance.end_date.isoformat() if set_dates else now_isoformat,
            'duration': task_instance.duration or 0,
            'attempt_number': task_instance.prev_attempted_tries,
            "payload": EqVariable("payload"),
        }
        expected_exception_message = truncate_string(
            exception_message,
            _EXCEPTION_MSG_LIMIT,
        ) if exception_message else None

        if exception_message:
            expected_graphql_payload["exception_message"] = expected_exception_message

        mock_client_upload_result.assert_called()
        mock_client_upload_result.assert_called_with(
            AirflowEventsClient._UPLOAD_AIRFLOW_TASK_RESULT_OPERATION,
            expected_graphql_payload,
            {
                "dag_id": dag.dag_id,
                "env": self._get_env(),
                "run_id": dag_context["run_id"],
                "success": state == "success",
                "task": self._get_dag_task_instance_result(
                    task_instance,
                    expected_exception_message,
                    set_dates=set_dates
                ),
                "event_type": "task",
            }
        )

    @staticmethod
    def _create_task_instance(
            dag_id: str,
            task_id: str,
            state: str,
            running: bool = False,
            set_dates: bool = True,
    ) -> TaskInstance:
        task_instance = create_autospec(TaskInstance)
        task_instance.next_retry_datetime.return_value = None
        task_instance.inlets = []
        task_instance.outlets = []
        task_instance.task_id = task_id
        task_instance.state = state
        task_instance.log_url = f"http://airflow.com/{dag_id}/{task_instance.task_id}/log"
        task_instance.prev_attempted_tries = 0
        task_instance.duration = 10.5 if not running else None
        if set_dates:
            task_instance.execution_date = datetime.now(tz=timezone.utc) - timedelta(seconds=10)
            task_instance.start_date = datetime.now(tz=timezone.utc) - timedelta(seconds=9)
            task_instance.end_date = datetime.now(tz=timezone.utc)
        else:
            task_instance.execution_date = None
            task_instance.start_date = None
            task_instance.end_date = None
        task_instance.max_tries = 3
        task_instance.try_number = 1
        return task_instance

    @staticmethod
    def _get_dag_task_instance_result(
            task_instance: TaskInstance,
            exception_message: Optional[str] = None,
            set_dates: bool = True,
    ) -> Dict:
        now_isoformat = datetime.now(tz=timezone.utc).isoformat()
        return {
            "task_id": task_instance.task_id,
            "state": task_instance.state,
            "log_url": task_instance.log_url,
            "prev_attempted_tries": task_instance.prev_attempted_tries,
            "duration": task_instance.duration or 0,
            "execution_date": task_instance.execution_date.isoformat() if set_dates else now_isoformat,
            "start_date": task_instance.start_date.isoformat() if set_dates else now_isoformat,
            "end_date": task_instance.end_date.isoformat() if set_dates else now_isoformat,
            "next_retry_datetime": None,
            "max_tries": task_instance.max_tries,
            "try_number": task_instance.try_number,
            "exception_message": exception_message,
            "inlets": [],
            "outlets": [],
            "original_dates": ANY,
        }




from typing import Optional
from unittest import TestCase
from unittest.mock import patch, ANY, Mock

from airflow import AirflowException
from airflow.models import Connection
from pycarlo.core import Session, Client

from airflow_mcd.hooks import SessionHook

SAMPLE_ID = 'foo'
SAMPLE_TOKEN = 'bar'
SAMPLE_CONN_ID = 'mcd_default_session'


class SessionHookTest(TestCase):
    def setUp(self) -> None:
        self._session = SessionHook(mcd_session_conn_id=SAMPLE_CONN_ID)

    def test_session_id_is_set(self):
        self.assertEqual(self._session.mcd_session_conn_id, SAMPLE_CONN_ID)

    @patch('airflow_mcd.hooks.session_hook.Session')
    @patch.object(SessionHook, 'get_connection')
    def test_get_conn_with_extra(self, get_connection_mock, session_mock):
        get_connection_mock.return_value = Connection(extra={
            'mcd_id': SAMPLE_ID,
            'mcd_token': SAMPLE_TOKEN
        })
        expected_session = Session(mcd_id=SAMPLE_ID, mcd_token=SAMPLE_TOKEN)
        session_mock.return_value = expected_session

        self.assertEqual(self._session.get_conn(), expected_session)
        get_connection_mock.assert_called_once_with(SAMPLE_CONN_ID)
        session_mock.assert_called_once_with(mcd_id=SAMPLE_ID, mcd_token=SAMPLE_TOKEN)

    @patch('airflow_mcd.hooks.session_hook.Session')
    @patch.object(SessionHook, 'get_connection')
    def test_get_conn_with_login_and_password(self, get_connection_mock, session_mock):
        get_connection_mock.return_value = Connection(login=SAMPLE_ID, password=SAMPLE_TOKEN)
        expected_session = Session(mcd_id=SAMPLE_ID, mcd_token=SAMPLE_TOKEN)
        session_mock.return_value = expected_session

        self.assertEqual(self._session.get_conn(), expected_session)
        get_connection_mock.assert_called_once_with(SAMPLE_CONN_ID)
        session_mock.assert_called_once_with(mcd_id=SAMPLE_ID, mcd_token=SAMPLE_TOKEN)

    @patch('airflow_mcd.hooks.session_hook.Session')
    @patch.object(SessionHook, 'get_connection')
    def test_get_conn_with_non_default_host(self, get_connection_mock, session_mock):
        host = 'http://localhost:8080'
        get_connection_mock.return_value = Connection(host=host, login=SAMPLE_ID, password=SAMPLE_TOKEN)
        expected_session = Session(mcd_id=SAMPLE_ID, mcd_token=SAMPLE_TOKEN)
        session_mock.return_value = expected_session

        self.assertEqual(self._session.get_conn(), expected_session)
        get_connection_mock.assert_called_once_with(SAMPLE_CONN_ID)
        session_mock.assert_called_once_with(
            endpoint=f'{host}/graphql',
            mcd_id=SAMPLE_ID,
            mcd_token=SAMPLE_TOKEN
        )

    @patch('airflow_mcd.hooks.session_hook.Session')
    @patch.object(SessionHook, 'get_connection')
    def test_get_conn_with_non_default_host_including_path(self, get_connection_mock, session_mock):
        host = 'http://localhost:8080/graphql'
        get_connection_mock.return_value = Connection(host=host, login=SAMPLE_ID, password=SAMPLE_TOKEN)
        expected_session = Session(mcd_id=SAMPLE_ID, mcd_token=SAMPLE_TOKEN)
        session_mock.return_value = expected_session

        self.assertEqual(self._session.get_conn(), expected_session)
        get_connection_mock.assert_called_once_with(SAMPLE_CONN_ID)
        session_mock.assert_called_once_with(
            endpoint=host,
            mcd_id=SAMPLE_ID,
            mcd_token=SAMPLE_TOKEN
        )

    @patch.object(SessionHook, 'get_connection')
    def test_get_conn_with_incomplete_config(self, get_connection_mock):
        get_connection_mock.return_value = Connection()

        with self.assertRaises(AirflowException) as context:
            self._session.get_conn()
        self.assertEqual(str(context.exception), 'Missing expected key \'mcd_id\' from connection extra.')

    @patch("airflow_mcd.hooks.session_hook.Session")
    @patch.object(SessionHook, "get_connection")
    @patch("requests.request")
    def test_get_gw_conn(self, mock_request, get_connection_mock, session_mock):
        host = "https://integrations.getmontecarlo.com"
        # when path is not specified, we initialize the session with /graphql, but that's ignored
        # later in make_request
        self._test_gw_conn(
            mock_request,
            get_connection_mock,
            session_mock,
            host,
            f"{host}/graphql"
        )

    @patch("airflow_mcd.hooks.session_hook.Session")
    @patch.object(SessionHook, "get_connection")
    @patch("requests.request")
    def test_get_gw_conn_trailing_slash(self, mock_request, get_connection_mock, session_mock):
        self._test_gw_conn(
            mock_request,
            get_connection_mock,
            session_mock,
            "https://integrations.getmontecarlo.com/"
        )

    @patch("airflow_mcd.hooks.session_hook.Session")
    @patch.object(SessionHook, "get_connection")
    @patch("requests.request")
    def test_get_gw_conn_any_path(self, mock_request, get_connection_mock, session_mock):
        self._test_gw_conn(
            mock_request,
            get_connection_mock,
            session_mock,
            "https://integrations.getmontecarlo.com/ignored/path"
        )

    def _test_gw_conn(
        self,
        mock_request: Mock,
        get_connection_mock: Mock,
        session_mock: Mock,
        host: str,
        expected_endpoint: Optional[str] = None,
    ):
        expected_endpoint = expected_endpoint or host
        get_connection_mock.return_value = Connection(
            host=host, login=SAMPLE_ID, password=SAMPLE_TOKEN
        )
        expected_session = Session(mcd_id=SAMPLE_ID, mcd_token=SAMPLE_TOKEN, endpoint=host, scope="AirflowCallbacks")
        session_mock.return_value = expected_session

        self.assertEqual(self._session.get_conn(), expected_session)
        get_connection_mock.assert_called_once_with(SAMPLE_CONN_ID)
        session_mock.assert_called_once_with(
            endpoint=expected_endpoint, mcd_id=SAMPLE_ID, mcd_token=SAMPLE_TOKEN, scope="AirflowCallbacks"
        )

        client = Client(self._session.get_conn())
        client.make_request("/airflow/callbacks", body={})
        mock_request.assert_called_once_with(
            url='https://integrations.getmontecarlo.com/airflow/callbacks',
            method='POST',
            json={},
            headers={
                "x-mcd-id": "foo",
                "x-mcd-token": "bar",
                "x-mcd-session-id": ANY,
                "x-mcd-trace-id": ANY,
                "x-mcd-telemetry-reason": "user",
            },
            timeout=10,
        )



from abc import ABC

from airflow.models import BaseOperator
from pycarlo.core import Session

from airflow_mcd.hooks.session_hook import SessionHook


class BaseMcdOperator(BaseOperator, ABC):
    def __init__(self, mcd_session_conn_id: str, *args, **kwargs):
        """
        Base Operator for MCD.

        :param mcd_session_conn_id: Connection ID for the MCD session.
        """
        super().__init__(*args, **kwargs)

        self.mcd_session_conn_id = mcd_session_conn_id

    def get_session(self) -> Session:
        """
        Gets session from hook.

        :return: MCD access session.
        """
        return SessionHook(mcd_session_conn_id=self.mcd_session_conn_id).get_conn()



from typing import Dict, Union, Optional
from uuid import UUID

from airflow import AirflowException
from pycarlo.core import Client
from pycarlo.features.circuit_breakers import CircuitBreakerService

from airflow_mcd.operators.base_operator import BaseMcdOperator

try:
    # Use AirflowFailException if available as that is preferred for tasks that can be failed without retrying.
    # https://airflow.apache.org/docs/apache-airflow/stable/_api/airflow/exceptions/index.html#airflow.exceptions.AirflowFailException
    from airflow.exceptions import AirflowFailException, AirflowSkipException

    BREACH_EXCEPTION = AirflowFailException
except:
    BREACH_EXCEPTION = AirflowException
    OP_RETRIES = 0
else:
    OP_RETRIES = None


class SimpleCircuitBreakerOperator(BaseMcdOperator):
    def __init__(self,
                 mcd_session_conn_id: str,
                 rule_uuid: Optional[Union[str, UUID]] = None,
                 namespace: Optional[str] = None,
                 rule_name: Optional[str] = None,
                 timeout_in_minutes: Optional[int] = 5,
                 fail_open: Optional[bool] = True,
                 runtime_variables: Optional[Dict[str, str]] = None,
                 *args, **kwargs):
        """
        Operator for Circuit breaking rules (custom SQL monitors) with MCD.

        Raises an AirflowFailException if the rule condition is in breach when using an Airflow version newer
        than 1.10.11, as that is preferred for tasks that can be failed without retrying.

        Older Airflow versions raise an AirflowException In this case retries are also set to zero as this operator is
        not intended to be retried.

        :param mcd_session_conn_id: Connection ID for the MCD session.
        :param rule_uuid: UUID of the rule (custom SQL monitor) to execute. Either the rule_uuid or the rule_name
        must be specified
        :param namespace: Namespace of the rule (custom SQL monitor) to execute.
        :param rule_name: name of the rule (custom SQL monitor) to execute.
        :param timeout_in_minutes: Polling timeout in minutes. Note that The Data Collector Lambda has a max timeout of
        15 minutes when executing a query. Queries that take longer to execute are not supported, so we recommend
        filtering down the query output to improve performance (e.g limit WHERE clause). If you expect a query to
        take the full 15 minutes we recommend padding the timeout to 20 minutes.
        :param fail_open: Prevent any errors or timeouts when executing a rule from stopping your pipeline.
        :param runtime_variables: Runtime variables to pass to the rule.
        Raises AirflowSkipException if set to True.
        """
        retry_handler = dict(retries=OP_RETRIES) if OP_RETRIES is not None else {}
        super().__init__(mcd_session_conn_id=mcd_session_conn_id, *args, **kwargs, **retry_handler)

        self.rule_uuid = rule_uuid
        self.namespace = namespace
        self.rule_name = rule_name
        self.timeout_in_minutes = timeout_in_minutes
        self.fail_open = fail_open
        self.runtime_variables = runtime_variables

    def execute(self, *args, **kwargs) -> None:
        """
        Execute the circuit breaker operator.
        """
        service = CircuitBreakerService(mc_client=Client(self.get_session()), print_func=self.log.info)
        try:
            in_breach = service.trigger_and_poll(rule_uuid=self.rule_uuid,
                                                 namespace=self.namespace,
                                                 rule_name=self.rule_name,
                                                 timeout_in_minutes=self.timeout_in_minutes,
                                                 runtime_variables=self.runtime_variables)
        except Exception as err:
            if not self.fail_open:
                raise AirflowException from err

            message = 'Encountered an error when executing the rule, but failing open.'
            self.log.exception(message)
            raise AirflowSkipException(message)
        else:
            if in_breach:
                raise BREACH_EXCEPTION(f'Rule \'{self.rule_uuid}\' is in breach!')
            self.log.info(f'Rule \'{self.rule_uuid}\' is not in breach.')



import logging
from copy import deepcopy

from pycarlo.common.utils import truncate_string
from airflow.models import TaskInstance, SlaMiss, DagRun, DAG
from attr import asdict
from datetime import datetime, timezone

from airflow_mcd.callbacks.client import DagResult, DagTaskResult, AirflowEventsClient, DagTaskInstanceResult, \
    SlaMissesResult, TaskSlaMiss
from typing import Dict, List, Any, Optional


logger = logging.getLogger(__name__)

_DEFAULT_CALL_TIMEOUT = 10
_SUCCESS_STATES = ['success', 'skipped']
_EXCEPTION_MSG_LIMIT = 10 * 1024  # 10kb

_DEFAULT_CONNECTION_ID = "mcd_gateway_default_session"
_FALLBACK_CONNECTION_ID = "mcd_default_session"


class AirflowEventsClientUtils:
    @classmethod
    def mcd_post_dag_result(
            cls,
            context: Dict,
    ):
        if not cls._validate_dag_callback_context(context=context):
            return

        dag: DAG = context['dag']
        dag_run: DagRun = context['dag_run']
        task_instances = dag_run.get_task_instances()
        result = DagResult(
            dag_id=dag.dag_id,
            run_id=context['run_id'],
            success=dag_run.state in _SUCCESS_STATES,
            reason=context['reason'],
            tasks=[cls._get_task_instance_result(ti) for ti in task_instances],
            state=dag_run.state,
            execution_date=cls._get_datetime_isoformat(dag_run.execution_date),
            start_date=cls._get_datetime_isoformat(dag_run.start_date),
            end_date=cls._get_datetime_isoformat(dag_run.end_date),
            original_dates=cls._get_original_dates(dag_run.execution_date, dag_run.start_date, dag_run.end_date),
        )
        cls._get_events_client(dag).upload_dag_result(result)

    @staticmethod
    def _validate_dag_callback_context(context: Dict) -> bool:
        error_message: Optional[str] = None
        if 'dag' not in context or 'run_id' not in context or 'dag_run' not in context:
            error_message = 'dag, run_id and dag_run are expected'
        else:
            dag_run: DagRun = context['dag_run']
            if not dag_run.end_date:
                error_message = 'no dag_run.end_date set, it looks like the dag is still running'
            elif 'reason' not in context:
                error_message = 'no reason set, it looks like the dag is still running'

        if error_message:
            logger.error(f'Invalid context received in MCD dag callback: {error_message}. '
                         'Please check your callbacks are configured properly.')
            return False
        return True

    @classmethod
    def mcd_post_task_result(cls, context: Dict):
        if 'dag' not in context or 'run_id' not in context or 'task_instance' not in context:
            logger.error('Invalid context received in MCD task callback: dag, run_id and task_instance are expected')
            return

        dag = context['dag']
        ti = context['task_instance']
        exception_message = truncate_string(
            str(context['exception']),
            _EXCEPTION_MSG_LIMIT,
        ) if 'exception' in context else None
        task_instance_result = cls._get_task_instance_result(ti, exception_message)

        result = DagTaskResult(
            dag_id=dag.dag_id,
            run_id=context['run_id'],
            success=task_instance_result.state in _SUCCESS_STATES,
            task=task_instance_result,
        )
        cls._get_events_client(dag).upload_task_result(result)

    @classmethod
    def mcd_post_sla_misses(cls, dag: DAG, sla_misses: List[SlaMiss]):
        result = SlaMissesResult(
            dag_id=dag.dag_id,
            sla_misses=[
                TaskSlaMiss(
                    task_id=sla_miss.task_id,
                    execution_date=cls._get_datetime_isoformat(sla_miss.execution_date),
                    timestamp=cls._get_datetime_isoformat(sla_miss.timestamp),
                ) for sla_miss in sla_misses
            ]
        )
        cls._get_events_client(dag).upload_sla_misses(result)

    @classmethod
    def _get_task_instance_result(
            cls,
            ti: TaskInstance,
            exception_message: Optional[str] = None
    ) -> DagTaskInstanceResult:
        return DagTaskInstanceResult(
            task_id=ti.task_id,
            state=ti.state,
            log_url=ti.log_url,
            prev_attempted_tries=ti.prev_attempted_tries,
            duration=ti.duration or 0,
            execution_date=cls._get_datetime_isoformat(ti.execution_date),
            start_date=cls._get_datetime_isoformat(ti.start_date),
            end_date=cls._get_datetime_isoformat(ti.end_date),
            next_retry_datetime=cls._get_next_retry_datetime(ti),
            max_tries=ti.max_tries,
            try_number=ti.try_number,
            exception_message=exception_message,
            inlets=cls._get_lineage_list(ti, 'inlets'),
            outlets=cls._get_lineage_list(ti, 'outlets'),
            original_dates=cls._get_original_dates(ti.execution_date, ti.start_date, ti.end_date),
        )

    @staticmethod
    def _get_datetime_isoformat(d: Optional[datetime]) -> str:
        return d.isoformat() if d else datetime.now(tz=timezone.utc).isoformat()

    @staticmethod
    def _get_original_dates(
            execution_date: Optional[datetime],
            start_date: Optional[datetime],
            end_date: Optional[datetime]
    ) -> str:
        return f"execution={str(execution_date)}, start_date={str(start_date)}, end_date={str(end_date)}"

    @staticmethod
    def _get_optional_datetime_isoformat(d: Optional[datetime]) -> Optional[str]:
        return d.isoformat() if d else None

    @classmethod
    def _get_events_client(cls, dag: DAG) -> AirflowEventsClient:
        dag_params = dag.params or {}
        mcd_session_conn_id = _DEFAULT_CONNECTION_ID
        mcd_fallback_conn_id: Optional[str] = _FALLBACK_CONNECTION_ID

        param_value = dag_params.get('mcd_connection_id')
        # in Airflow 2.2.x we're getting a Param object while in Airflow 2.6.x we're getting a string object
        # but we cannot import Param as it was added in Airflow v2 and not present in Airflow v1
        if hasattr(param_value, "value") and isinstance(param_value.value, str):
            mcd_session_conn_id = param_value.value
        elif isinstance(param_value, str):
            mcd_session_conn_id = param_value
        elif param_value is not None:  # don't log a warning when the parameter was not specified at all
            logger.warning(f"Ignoring mcd_connection_id parameter value: {param_value}, using {mcd_session_conn_id}")

        if mcd_session_conn_id != _DEFAULT_CONNECTION_ID:
            # don't fallback to the old connection id when the value was specifically set
            mcd_fallback_conn_id = None

        return AirflowEventsClient(
            mcd_session_conn_id=mcd_session_conn_id,
            mcd_fallback_conn_id=mcd_fallback_conn_id,
            call_timeout=_DEFAULT_CALL_TIMEOUT,
        )

    @classmethod
    def _get_next_retry_datetime(cls, ti: TaskInstance) -> Optional[str]:
        if not hasattr(ti, "task") or not ti.task or not ti.end_date:
            return None
        return cls._get_optional_datetime_isoformat(ti.next_retry_datetime())

    @classmethod
    def _get_lineage_dict(cls, o: Any) -> Dict:
        attrs = deepcopy(asdict(o))
        attrs['type'] = str(type(o))
        return attrs

    @classmethod
    def _get_lineage_list(cls, ti: TaskInstance, attr: str) -> Optional[List[Dict]]:
        if not hasattr(ti, "task") or not ti.task:
            return None
        lineage_list = getattr(ti.task, attr)
        if not lineage_list:
            return None
        return [
            cls._get_lineage_dict(lineage_object) for lineage_object in lineage_list
        ]



import os
from unittest import TestCase
from unittest.mock import patch

from airflow_mcd.operators.dbt import DbtConfig, DefaultConfigProvider


class TestConfigProvider(DefaultConfigProvider):

    def config(self) -> DbtConfig:
        return DbtConfig(dbt_bin="/usr/bin/dbt")


class InvalidConfigProvider:

    @staticmethod
    def config() -> DbtConfig:
        return DbtConfig(dbt_bin="/usr/bin/dbt")


class DefaultConfigProviderTests(TestCase):

    def test_default_config(self):
        # when
        config = DefaultConfigProvider.get()

        # then
        self.assertEqual("dbt", config.dbt_bin)

    @patch.dict(
        os.environ,
        {
            DefaultConfigProvider.ENV_VAR_NAME: "tests.operators.test_dbt_default_config_provider.TestConfigProvider"
        },
        clear=True
    )
    def test_custom_config(self):
        # when
        config = DefaultConfigProvider.get()

        # then
        self.assertEqual("/usr/bin/dbt", config.dbt_bin)

    @patch.dict(
        os.environ,
        {
            f"AIRFLOW__MC__{DefaultConfigProvider.ENV_VAR_NAME}": "tests.operators.test_dbt_default_config_provider.TestConfigProvider"
        },
        clear=True
    )
    def test_custom_config_mwaa(self):
        # when
        config = DefaultConfigProvider.get()

        # then
        self.assertEqual("/usr/bin/dbt", config.dbt_bin)


    @patch.dict(
        os.environ,
        {
            DefaultConfigProvider.ENV_VAR_NAME: "tests.operators.test_dbt_default_config_provider.InvalidConfigProvider"
        },
        clear=True
    )
    def test_invalid_config(self):
        # when
        with self.assertRaises(Exception) as context:
            DefaultConfigProvider.get()

        # then
        self.assertEqual(
            "dbt configuration provider defined in AIRFLOW_MCD_DBT_CONFIG_PROVIDER is not an instance of DefaultConfigProvider",
            str(context.exception)
        )



import os
import subprocess
from unittest import TestCase
from unittest.mock import Mock, patch

from airflow.exceptions import AirflowException
from airflow.utils.context import Context

from airflow_mcd.operators.dbt import DbtOperator


class DbtOperatorTests(TestCase):

    @patch("airflow_mcd.operators.dbt.subprocess")
    @patch("airflow_mcd.operators.dbt.open")
    @patch("airflow_mcd.operators.dbt.DbtImporter")
    @patch("airflow_mcd.operators.dbt.Client")
    @patch("airflow_mcd.operators.dbt.SessionHook")
    @patch.dict(os.environ, {}, clear=True)
    def test_operator_with_minimal_args(
            self,
            mock_mc_session_hook: Mock,
            mock_mc_client: Mock,
            mock_dbt_importer: Mock,
            mock_open: Mock,
            mock_subprocess: Mock,
    ):
        # given
        mock_stdout = Mock()
        mock_stdout.readline.side_effect = [b"logging...\n"]
        mock_process = Mock()
        mock_process.returncode = 0
        mock_process.stdout = mock_stdout
        mock_subprocess.Popen.return_value = mock_process
        mock_subprocess.PIPE = subprocess.PIPE
        mock_subprocess.STDOUT = subprocess.STDOUT

        mock_log_file = Mock()
        mock_open.return_value = mock_log_file

        mock_mc_session = Mock()
        mock_mc_session_hook_instance = Mock()
        mock_mc_session_hook_instance.get_conn.return_value = mock_mc_session
        mock_mc_session_hook.return_value = mock_mc_session_hook_instance

        mock_mc_client_instance = Mock()
        mock_mc_client.return_value = mock_mc_client_instance

        mock_dbt_importer_instance = Mock()
        mock_dbt_importer.return_value = mock_dbt_importer_instance

        operator = DbtOperator(
            task_id="test-operator",
            command="run",
            project_name="test-project",
            job_name="test-job",
            mc_conn_id="mc-connection",
            mc_include_logs=True,
        )

        # when
        operator.execute(Context())

        # verify expected call to submit dbt command
        mock_subprocess.Popen.assert_called_once_with(
            ["dbt", "run"],
            env={},
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            cwd=".",
            close_fds=True,
        )

        # verify expected call to open local log file
        mock_open.assert_called_once_with(
            "./dbt.log",
            "w",
        )

        # verify expected calls to write to local log file
        mock_log_file.write.assert_called_once_with("logging...\n")
        mock_log_file.close.assert_called_once_with()

        # verify expected call to wait for dbt command to complete
        mock_process.wait.assert_called_once_with()

        # verify expected call to create MC SessionHook
        mock_mc_session_hook.assert_called_once_with(
            mcd_session_conn_id="mc-connection"
        )

        # verify expected call to create MC API client
        mock_mc_client.assert_called_once_with(
            session=mock_mc_session
        )

        # verify expected call to create MC DbtImporter
        mock_dbt_importer.assert_called_once_with(
            mc_client=mock_mc_client_instance
        )

        # verify expected call to import dbt run results to MC
        mock_dbt_importer_instance.import_run.assert_called_once_with(
            manifest_path="./target/manifest.json",
            run_results_path="./target/run_results.json",
            logs_path="./dbt.log",
            project_name="test-project",
            job_name="test-job",
            resource_id=None,
        )

    @patch("airflow_mcd.operators.dbt.subprocess")
    @patch("airflow_mcd.operators.dbt.open")
    @patch("airflow_mcd.operators.dbt.DbtImporter")
    @patch("airflow_mcd.operators.dbt.Client")
    @patch("airflow_mcd.operators.dbt.SessionHook")
    @patch.dict(os.environ, {}, clear=True)
    def test_operator_with_full_args(
            self,
            mock_mc_session_hook: Mock,
            mock_mc_client: Mock,
            mock_dbt_importer: Mock,
            mock_open: Mock,
            mock_subprocess: Mock,
    ):
        # given
        mock_stdout = Mock()
        mock_stdout.readline.side_effect = [b"logging...\n"]
        mock_process = Mock()
        mock_process.returncode = 0
        mock_process.stdout = mock_stdout
        mock_subprocess.Popen.return_value = mock_process
        mock_subprocess.PIPE = subprocess.PIPE
        mock_subprocess.STDOUT = subprocess.STDOUT

        mock_log_file = Mock()
        mock_open.return_value = mock_log_file

        mock_mc_session = Mock()
        mock_mc_session_hook_instance = Mock()
        mock_mc_session_hook_instance.get_conn.return_value = mock_mc_session
        mock_mc_session_hook.return_value = mock_mc_session_hook_instance

        mock_mc_client_instance = Mock()
        mock_mc_client.return_value = mock_mc_client_instance

        mock_dbt_importer_instance = Mock()
        mock_dbt_importer.return_value = mock_dbt_importer_instance

        operator = DbtOperator(
            task_id="test-operator",
            command="run",
            project_name="test-project",
            job_name="test-job",
            mc_conn_id="mc-connection",
            mc_resource_id="resource-id",
            mc_include_logs=True,
            dir="/usr/local/dbt/project",
            dbt_bin="/usr/bin/dbt",
            env={"TEST": "ING"},
            profiles_dir="/usr/local/dbt/profiles",
            target_path="/usr/local/dbt/target",
            target="prod",
            vars={"foo": "bar"},
            models="some_model",
            exclude="excluded_model",
            select="selected_model",
            warn_error=True,
            full_refresh=True,
            data=True,
            schema=True,
        )

        # when
        operator.execute(Context())

        # verify expected call to submit dbt command
        mock_subprocess.Popen.assert_called_once_with(
            [
                "/usr/bin/dbt",
                "--warn-error",
                "run",
                "--profiles-dir",
                "/usr/local/dbt/profiles",
                "--target",
                "prod",
                "--vars",
                "{\"foo\": \"bar\"}",
                "--models",
                "some_model",
                "--exclude",
                "excluded_model",
                "--select",
                "selected_model",
                "--full-refresh",
                "--data",
                "--schema",
            ],
            env={
                "TEST": "ING",
            },
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            cwd="/usr/local/dbt/project",
            close_fds=True,
        )

        # verify expected call to open local log file
        mock_open.assert_called_once_with(
            "/usr/local/dbt/project/dbt.log",
            "w",
        )

        # verify expected calls to write to local log file
        mock_log_file.write.assert_called_once_with("logging...\n")
        mock_log_file.close.assert_called_once_with()

        # verify expected call to wait for dbt command to complete
        mock_process.wait.assert_called_once_with()

        # verify expected call to create MC SessionHook
        mock_mc_session_hook.assert_called_once_with(
            mcd_session_conn_id="mc-connection"
        )

        # verify expected call to create MC API client
        mock_mc_client.assert_called_once_with(
            session=mock_mc_session
        )

        # verify expected call to create MC DbtImporter
        mock_dbt_importer.assert_called_once_with(
            mc_client=mock_mc_client_instance
        )

        # verify expected call to import dbt run results to MC
        mock_dbt_importer_instance.import_run.assert_called_once_with(
            manifest_path="/usr/local/dbt/target/manifest.json",
            run_results_path="/usr/local/dbt/target/run_results.json",
            logs_path="/usr/local/dbt/project/dbt.log",
            project_name="test-project",
            job_name="test-job",
            resource_id="resource-id",
        )

    @patch("airflow_mcd.operators.dbt.subprocess")
    @patch("airflow_mcd.operators.dbt.open")
    @patch("airflow_mcd.operators.dbt.DbtImporter")
    @patch("airflow_mcd.operators.dbt.Client")
    @patch("airflow_mcd.operators.dbt.SessionHook")
    @patch.dict(os.environ, {}, clear=True)
    def test_operator_with_failed_dbt_command(
            self,
            mock_mc_session_hook: Mock,
            mock_mc_client: Mock,
            mock_dbt_importer: Mock,
            mock_open: Mock,
            mock_subprocess: Mock,
    ):
        # given
        mock_stdout = Mock()
        mock_stdout.readline.side_effect = [b"logging...\n"]
        mock_process = Mock()
        mock_process.returncode = 1
        mock_process.stdout = mock_stdout
        mock_subprocess.Popen.return_value = mock_process
        mock_subprocess.PIPE = subprocess.PIPE
        mock_subprocess.STDOUT = subprocess.STDOUT

        mock_log_file = Mock()
        mock_open.return_value = mock_log_file

        mock_mc_session = Mock()
        mock_mc_session_hook_instance = Mock()
        mock_mc_session_hook_instance.get_conn.return_value = mock_mc_session
        mock_mc_session_hook.return_value = mock_mc_session_hook_instance

        mock_mc_client_instance = Mock()
        mock_mc_client.return_value = mock_mc_client_instance

        mock_dbt_importer_instance = Mock()
        mock_dbt_importer.return_value = mock_dbt_importer_instance

        operator = DbtOperator(
            task_id="test-operator",
            command="run",
            project_name="test-project",
            job_name="test-job",
            mc_conn_id="mc-connection",
            mc_include_logs=True,
        )

        # when
        with self.assertRaises(AirflowException) as context:
            operator.execute(Context())

        # verify expected exception
        self.assertEqual("dbt command failed", str(context.exception))

        # verify expected call to submit dbt command
        mock_subprocess.Popen.assert_called_once_with(
            ["dbt", "run"],
            env={},
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            cwd=".",
            close_fds=True,
        )

        # verify expected call to open local log file
        mock_open.assert_called_once_with(
            "./dbt.log",
            "w",
        )

        # verify expected calls to write to local log file
        mock_log_file.write.assert_called_once_with("logging...\n")
        mock_log_file.close.assert_called_once_with()

        # verify expected call to wait for dbt command to complete
        mock_process.wait.assert_called_once_with()

        # verify expected call to create MC SessionHook
        mock_mc_session_hook.assert_called_once_with(
            mcd_session_conn_id="mc-connection"
        )

        # verify expected call to create MC API client
        mock_mc_client.assert_called_once_with(
            session=mock_mc_session
        )

        # verify expected call to create MC DbtImporter
        mock_dbt_importer.assert_called_once_with(
            mc_client=mock_mc_client_instance
        )

        # verify expected call to import dbt run results to MC
        mock_dbt_importer_instance.import_run.assert_called_once_with(
            manifest_path="./target/manifest.json",
            run_results_path="./target/run_results.json",
            logs_path="./dbt.log",
            project_name="test-project",
            job_name="test-job",
            resource_id=None,
        )


    @patch("airflow_mcd.operators.dbt.subprocess")
    @patch("airflow_mcd.operators.dbt.open")
    @patch("airflow_mcd.operators.dbt.DbtImporter")
    @patch("airflow_mcd.operators.dbt.Client")
    @patch("airflow_mcd.operators.dbt.SessionHook")
    @patch.dict(os.environ, {}, clear=True)
    def test_operator_with_mc_disabled(
            self,
            mock_mc_session_hook: Mock,
            mock_mc_client: Mock,
            mock_dbt_importer: Mock,
            mock_open: Mock,
            mock_subprocess: Mock,
    ):
        # given
        mock_stdout = Mock()
        mock_stdout.readline.side_effect = [b"logging...\n"]
        mock_process = Mock()
        mock_process.returncode = 0
        mock_process.stdout = mock_stdout
        mock_subprocess.Popen.return_value = mock_process
        mock_subprocess.PIPE = subprocess.PIPE
        mock_subprocess.STDOUT = subprocess.STDOUT

        mock_log_file = Mock()
        mock_open.return_value = mock_log_file

        operator = DbtOperator(
            task_id="test-operator",
            command="run",
            project_name="test-project",
            job_name="test-job",
            mc_conn_id="mc-connection",
            mc_include_logs=True,
            mc_enabled=False,
        )

        # when
        operator.execute(Context())

        # verify expected call to submit dbt command
        mock_subprocess.Popen.assert_called_once_with(
            ["dbt", "run"],
            env={},
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            cwd=".",
            close_fds=True,
        )

        # verify expected call to open local log file
        mock_open.assert_called_once_with(
            "./dbt.log",
            "w",
        )

        # verify expected calls to write to local log file
        mock_log_file.write.assert_called_once_with("logging...\n")
        mock_log_file.close.assert_called_once_with()

        # verify expected call to wait for dbt command to complete
        mock_process.wait.assert_called_once_with()

        # verify no calls to MC APIs
        mock_mc_session_hook.assert_not_called()
        mock_mc_client.assert_not_called()
        mock_dbt_importer.assert_not_called()

    @patch("airflow_mcd.operators.dbt.subprocess")
    @patch("airflow_mcd.operators.dbt.open")
    @patch("airflow_mcd.operators.dbt.DbtImporter")
    @patch("airflow_mcd.operators.dbt.Client")
    @patch("airflow_mcd.operators.dbt.SessionHook")
    @patch.dict(os.environ, {}, clear=True)
    def test_operator_with_failed_import_and_mc_success_not_required(
            self,
            mock_mc_session_hook: Mock,
            mock_mc_client: Mock,
            mock_dbt_importer: Mock,
            mock_open: Mock,
            mock_subprocess: Mock,
    ):
        # given
        mock_stdout = Mock()
        mock_stdout.readline.side_effect = [b"logging...\n"]
        mock_process = Mock()
        mock_process.returncode = 0
        mock_process.stdout = mock_stdout
        mock_subprocess.Popen.return_value = mock_process
        mock_subprocess.PIPE = subprocess.PIPE
        mock_subprocess.STDOUT = subprocess.STDOUT

        mock_log_file = Mock()
        mock_open.return_value = mock_log_file

        mock_mc_session = Mock()
        mock_mc_session_hook_instance = Mock()
        mock_mc_session_hook_instance.get_conn.return_value = mock_mc_session
        mock_mc_session_hook.return_value = mock_mc_session_hook_instance

        mock_mc_client_instance = Mock()
        mock_mc_client.return_value = mock_mc_client_instance

        mock_dbt_importer_instance = Mock()
        mock_dbt_importer_instance.import_run.side_effect = Exception("uh oh")
        mock_dbt_importer.return_value = mock_dbt_importer_instance

        operator = DbtOperator(
            task_id="test-operator",
            command="run",
            project_name="test-project",
            job_name="test-job",
            mc_conn_id="mc-connection",
            mc_include_logs=True,
        )

        # when (no exception should be raised)
        operator.execute(Context())

        # verify expected call to submit dbt command
        mock_subprocess.Popen.assert_called_once_with(
            ["dbt", "run"],
            env={},
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            cwd=".",
            close_fds=True,
        )

        # verify expected call to open local log file
        mock_open.assert_called_once_with(
            "./dbt.log",
            "w",
        )

        # verify expected calls to write to local log file
        mock_log_file.write.assert_called_once_with("logging...\n")
        mock_log_file.close.assert_called_once_with()

        # verify expected call to wait for dbt command to complete
        mock_process.wait.assert_called_once_with()

        # verify expected call to create MC SessionHook
        mock_mc_session_hook.assert_called_once_with(
            mcd_session_conn_id="mc-connection"
        )

        # verify expected call to create MC API client
        mock_mc_client.assert_called_once_with(
            session=mock_mc_session
        )

        # verify expected call to create MC DbtImporter
        mock_dbt_importer.assert_called_once_with(
            mc_client=mock_mc_client_instance
        )

        # verify expected call to import dbt run results to MC
        mock_dbt_importer_instance.import_run.assert_called_once_with(
            manifest_path="./target/manifest.json",
            run_results_path="./target/run_results.json",
            logs_path="./dbt.log",
            project_name="test-project",
            job_name="test-job",
            resource_id=None,
        )

    @patch("airflow_mcd.operators.dbt.subprocess")
    @patch("airflow_mcd.operators.dbt.open")
    @patch("airflow_mcd.operators.dbt.DbtImporter")
    @patch("airflow_mcd.operators.dbt.Client")
    @patch("airflow_mcd.operators.dbt.SessionHook")
    @patch.dict(os.environ, {}, clear=True)
    def test_operator_with_failed_import_and_mc_success_required(
            self,
            mock_mc_session_hook: Mock,
            mock_mc_client: Mock,
            mock_dbt_importer: Mock,
            mock_open: Mock,
            mock_subprocess: Mock,
    ):
        # given
        mock_stdout = Mock()
        mock_stdout.readline.side_effect = [b"logging...\n"]
        mock_process = Mock()
        mock_process.returncode = 0
        mock_process.stdout = mock_stdout
        mock_subprocess.Popen.return_value = mock_process
        mock_subprocess.PIPE = subprocess.PIPE
        mock_subprocess.STDOUT = subprocess.STDOUT

        mock_log_file = Mock()
        mock_open.return_value = mock_log_file

        mock_mc_session = Mock()
        mock_mc_session_hook_instance = Mock()
        mock_mc_session_hook_instance.get_conn.return_value = mock_mc_session
        mock_mc_session_hook.return_value = mock_mc_session_hook_instance

        mock_mc_client_instance = Mock()
        mock_mc_client.return_value = mock_mc_client_instance

        mock_dbt_importer_instance = Mock()
        mock_dbt_importer_instance.import_run.side_effect = Exception("uh oh")
        mock_dbt_importer.return_value = mock_dbt_importer_instance

        operator = DbtOperator(
            task_id="test-operator",
            command="run",
            project_name="test-project",
            job_name="test-job",
            mc_conn_id="mc-connection",
            mc_success_required=True,
            mc_include_logs=True,
        )

        # when
        with self.assertRaises(Exception) as context:
            operator.execute(Context())

        # verify expected exception
        self.assertEqual("uh oh", str(context.exception))

        # verify expected call to submit dbt command
        mock_subprocess.Popen.assert_called_once_with(
            ["dbt", "run"],
            env={},
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            cwd=".",
            close_fds=True,
        )

        # verify expected call to open local log file
        mock_open.assert_called_once_with(
            "./dbt.log",
            "w",
        )

        # verify expected calls to write to local log file
        mock_log_file.write.assert_called_once_with("logging...\n")
        mock_log_file.close.assert_called_once_with()

        # verify expected call to wait for dbt command to complete
        mock_process.wait.assert_called_once_with()

        # verify expected call to create MC SessionHook
        mock_mc_session_hook.assert_called_once_with(
            mcd_session_conn_id="mc-connection"
        )

        # verify expected call to create MC API client
        mock_mc_client.assert_called_once_with(
            session=mock_mc_session
        )

        # verify expected call to create MC DbtImporter
        mock_dbt_importer.assert_called_once_with(
            mc_client=mock_mc_client_instance
        )

        # verify expected call to import dbt run results to MC
        mock_dbt_importer_instance.import_run.assert_called_once_with(
            manifest_path="./target/manifest.json",
            run_results_path="./target/run_results.json",
            logs_path="./dbt.log",
            project_name="test-project",
            job_name="test-job",
            resource_id=None,
        )



from typing import Dict, Tuple, Optional
from urllib.parse import urlparse, urlunparse

from airflow.exceptions import AirflowException
from airflow.models.connection import Connection

from pycarlo.common.retries import ExponentialBackoff
from pycarlo.core import Session, Client, Query

from airflow_mcd.hooks.constants import (
    MCD_GATEWAY_CONNECTION_TYPE,
    MCD_GATEWAY_SCOPE,
    MCD_GATEWAY_HOSTS,
)

try:
    from airflow.hooks.base import BaseHook

    HOOK_SOURCE = None
except ImportError:
    # For Airflow 1.10.*
    from airflow.hooks.base_hook import BaseHook

    HOOK_SOURCE = 'mcd_session'
from pycarlo.core import Session

_DEFAULT_TIMEOUT_SEC = 10
_DEFAULT_BACKOFF_START_SEC = 2
_DEFAULT_BACKOFF_MAX_SEC = 90
_PERMISSION_TEST_QUERY = """
query getUser {
   getUser {
     auth{
      permissions {
        permission
        effect
      }
    }
   }
 }
"""
_REQUIRED_PERMISSIONS = ["CatalogEdit", "MonitorsAccess"]


class SessionHook(BaseHook):
    API_PATH = '/graphql'

    conn_name_attr = "mcd_session_conn_id"

    default_conn_name = "mcd_default_session"

    conn_type = "mcd"

    hook_name = "Monte Carlo Data"

    def __init__(self, mcd_session_conn_id: str):
        """
        MCD Session Hook. Retrieves connection details from the Airflow `Connection` object.

        The `mcd_id` can be configured via the connection "login", and the `mcd_token` via the connection "password".

        Alternatively, either `mcd_id` or `mcd_token` can be configured in the connection "extra", with values passed
        via "login" or "password" taking precedence.
        {
            "mcd_id": "foo",
            "mcd_token": "bar"
        }

        :param mcd_session_conn_id: Connection ID for the MCD session.
        """
        self.mcd_session_conn_id = mcd_session_conn_id

        super().__init__(**(dict(source=HOOK_SOURCE) if HOOK_SOURCE is not None else {}))

    def get_conn(self) -> Session:
        """
        Gets a connection for the hook.

        :return: MCD access session.
        """
        connection = self.get_connection(self.mcd_session_conn_id)
        connection_extra = connection.extra_dejson
        try:
            return Session(
                mcd_id=connection.login or connection_extra['mcd_id'],
                mcd_token=connection.password or connection_extra['mcd_token'],
                **self._get_session_extra(connection)
            )
        except KeyError as err:
            raise AirflowException(f'Missing expected key {err} from connection extra.')

    def _get_session_extra(self, connection: Connection) -> Dict:
        """
        Extract extra MCD session parameters from an Airflow connection.

        :param connection: Airflow connection
        :return: dictionary of kwargs for MCD Session
        """
        extras = {}
        if connection.host:
            extras['endpoint'] = self._get_api_endpoint(connection.host)
        if self._is_gateway_connection(connection.conn_type, connection.host):
            extras["scope"] = MCD_GATEWAY_SCOPE
        return extras

    @staticmethod
    def _is_gateway_connection(connection_type: str, host: Optional[str]) -> bool:
        if connection_type == MCD_GATEWAY_CONNECTION_TYPE:
            return True
        else:
            # in some environments our custom connection types (like "Monte Carlo Data Gateway") are not
            # showing up, as a workaround we're supporting HTTP connections with the endpoint being the
            # integration gateway
            return urlparse(host).netloc in MCD_GATEWAY_HOSTS

    def _get_api_endpoint(self, host: str) -> str:
        """
        Get MCD API endpoint from Airflow connection host.

        :param host: Airflow connection host
        :return: MCD API endpoint url
        """
        parsed = urlparse(host)
        return urlunparse((
            parsed.scheme,
            parsed.netloc,
            parsed.path or self.API_PATH,  # set root API path if not provided in connection host
            parsed.params,
            parsed.query,
            parsed.fragment
        ))

    def test_connection(self) -> Tuple[bool, str]:
        """
        Test MCD API connection.

        :return: tuple of boolean and string for whether the test was successful and a related message
        """
        try:
            mc_client = Client(self.get_conn())
            result = mc_client(
                query=_PERMISSION_TEST_QUERY,
                retry_backoff=ExponentialBackoff(_DEFAULT_BACKOFF_START_SEC, _DEFAULT_BACKOFF_MAX_SEC),
                timeout_in_seconds=_DEFAULT_TIMEOUT_SEC,
            )
            if "error" in result:
                return False, f"Failed to test MC connection: getUser query failed: {result.error.errors[0].message}"

            if not all(any(p.permission == required_permission and p.effect.lower() == "allow" for p in result.get_user.auth.permissions) for required_permission in _REQUIRED_PERMISSIONS):
                return False, f'Failed to test MC connection: Missing one of the required permissions: {_REQUIRED_PERMISSIONS}'

            return True, 'Connection successfully tested'
        except Exception as exc:
            return False, f'Failed to test MC connection: {exc}'



import os
import logging

from dataclasses import dataclass, field
from typing import Optional, List, Dict

import requests
from airflow.exceptions import AirflowNotFoundException
from airflow.hooks.base import BaseHook
from dataclasses_json import dataclass_json
from pycarlo.common.retries import ExponentialBackoff
from pycarlo.core import Session, Client, Mutation
from pycarlo.lib.schema import GenericScalar
from requests import HTTPError
from sgqlc.types import Variable, non_null

from airflow_mcd.hooks import SessionHook
from airflow_mcd.hooks.gateway_session_hook import GatewaySessionHook

logger = logging.getLogger(__name__)

_ENV_MAPPINGS = {
    'env_name': [
        'AIRFLOW_ENV_NAME',                     # AWS
        'COMPOSER_ENVIRONMENT',                 # GCP Composer
        'AIRFLOW__WEBSERVER__INSTANCE_NAME',    # Astronomer
    ],
    'env_id': [
        'AIRFLOW_ENV_ID',                       # AWS
        'COMPOSER_GKE_NAME',                    # GCP Composer
        'ASTRO_DEPLOYMENT_ID',                  # Astronomer
    ],
    'version': [
        'AIRFLOW_VERSION',                      # AWS
        'MAJOR_VERSION',                        # GCP Composer
        'ASTRONOMER_RUNTIME_VERSION',           # Astronomer
    ],
    'base_url': [
        'AIRFLOW__WEBSERVER__BASE_URL',         # available in all containers
    ]
}
_INTEGRATION_GATEWAY_PATH = "/airflow/callbacks"


def _get_env_value(key: str, default_value: Optional[str] = None) -> Optional[str]:
    for env_var in _ENV_MAPPINGS[key]:
        value = os.environ.get(env_var)
        if value:
            return value
    return default_value


@dataclass_json
@dataclass
class AirflowEnv:
    # these env vars are used to get additional information about the Airflow
    # environment when running in AWS, GCP or Astronomer.
    env_name: str = ""
    env_id: Optional[str] = None
    version: Optional[str] = None
    base_url: Optional[str] = None

    def __post_init__(self):
        if not self.env_name:
            self.env_name = _get_env_value('env_name', 'airflow')
        if not self.env_id:
            self.env_id = _get_env_value('env_id')
        if not self.version:
            self.version = _get_env_value('version')
        if not self.base_url:
            self.base_url = _get_env_value('base_url')


@dataclass_json
@dataclass
class DagTaskInstanceResult:
    task_id: str
    state: str
    log_url: str
    prev_attempted_tries: int
    duration: float
    execution_date: str
    start_date: str
    end_date: str
    next_retry_datetime: Optional[str]
    max_tries: int
    try_number: int
    exception_message: Optional[str]
    inlets: Optional[List[Dict]]
    outlets: Optional[List[Dict]]
    original_dates: Optional[str]


@dataclass_json
@dataclass
class BaseDagResult:
    dag_id: str
    env: AirflowEnv = field(init=False)

    def __post_init__(self):
        self.env = AirflowEnv()


@dataclass_json
@dataclass
class BaseDagRunResult(BaseDagResult):
    run_id: str
    success: bool


@dataclass_json
@dataclass
class DagResult(BaseDagRunResult):
    tasks: Optional[List[DagTaskInstanceResult]]
    state: str
    execution_date: str
    start_date: str
    end_date: str
    reason: Optional[str]
    event_type: str = 'dag'
    original_dates: Optional[str] = None


@dataclass_json
@dataclass
class DagTaskResult(BaseDagRunResult):
    task: DagTaskInstanceResult
    event_type: str = 'task'


@dataclass_json
@dataclass
class TaskSlaMiss:
    task_id: str
    execution_date: str
    timestamp: str


@dataclass_json
@dataclass
class SlaMissesResult(BaseDagResult):
    sla_misses: List[TaskSlaMiss]
    event_type: str = 'sla_miss'


class AirflowEventsClient:
    """
    Client class used to send Airflow related events to Monte Carlo.
    """

    _UPLOAD_AIRFLOW_DAG_RESULT_OPERATION = "upload_airflow_dag_result"
    _UPLOAD_AIRFLOW_TASK_RESULT_OPERATION = "upload_airflow_task_result"
    _UPLOAD_AIRFLOW_SLA_MISSES_OPERATION = "upload_airflow_sla_misses"

    def __init__(self, mcd_session_conn_id: str, call_timeout: int, mcd_fallback_conn_id: Optional[str] = None):
        self.mcd_session_conn_id = mcd_session_conn_id
        self.mcd_fallback_conn_id = mcd_fallback_conn_id
        self.mcd_call_timeout = call_timeout

    def upload_dag_result(self, result: DagResult):
        payload = {
            'dag_id': result.dag_id,
            'run_id': result.run_id,
            'success': result.success,
            'reason': result.reason,
            'state': result.state,
            'execution_date': result.execution_date,
            'start_date': result.start_date,
            'end_date': result.end_date,
            'env': self._env_to_payload(result.env),
            'payload': Variable('payload'),
        }
        self._upload_result(self._UPLOAD_AIRFLOW_DAG_RESULT_OPERATION, payload, result.to_dict())

    def upload_task_result(self, result: DagTaskResult):
        payload = {
            'dag_id': result.dag_id,
            'run_id': result.run_id,
            'task_id': result.task.task_id,
            'success': result.success,
            'state': result.task.state,
            'log_url': result.task.log_url,
            'execution_date': result.task.execution_date,
            'start_date': result.task.start_date,
            'end_date': result.task.end_date,
            'duration': result.task.duration,
            'attempt_number': result.task.prev_attempted_tries,
            'env': self._env_to_payload(result.env),
            'payload': Variable('payload'),
        }
        if result.task.next_retry_datetime:
            payload['next_retry_date'] = result.task.next_retry_datetime
        if result.task.exception_message:
            payload['exception_message'] = result.task.exception_message
        self._upload_result(self._UPLOAD_AIRFLOW_TASK_RESULT_OPERATION, payload, result.to_dict())

    def upload_sla_misses(self, result: SlaMissesResult):
        payload = {
            'dag_id': result.dag_id,
            'env': self._env_to_payload(result.env),
            'payload': Variable('payload'),
        }
        self._upload_result(self._UPLOAD_AIRFLOW_SLA_MISSES_OPERATION, payload, result.to_dict())

    def _upload_result(self, operation_name: str, operation_parameters: Dict, payload_value: Dict):
        conn_id = self._get_existing_connection_id()
        mc_client = Client(self._get_session(conn_id))
        if mc_client.session_scope:
            self._upload_result_igw(mc_client, operation_name, operation_parameters, payload_value)
        else:
            self._upload_result_gql(mc_client, operation_name, operation_parameters, payload_value)

    def _upload_result_gql(
        self,
        mc_client: Client,
        operation_name: str,
        operation_parameters: Dict,
        payload_value: Dict,
    ):
        try:
            query = Mutation(payload=non_null(GenericScalar))
            attr = getattr(query, operation_name)
            attr(**operation_parameters)
            mc_client(
                query=query,
                variables={'payload': payload_value},
                retry_backoff=ExponentialBackoff(0, 0),  # disable retries
                timeout_in_seconds=self.mcd_call_timeout,
            )
        except Exception as exc:
            logger.exception(f'Failed to upload information to MC: {exc}')

    def _should_retry(self, exc: Exception) -> bool:
        """
        Determines if the exception should be retried.
        We are retrying for:
        - 408: Request Timeout
        - 429: Too Many Requests
        - 5xx: Server errors

        NOTE: For 429 errors we should consider overriding the backoff strategy to avoid
        hammering the server when we are rate limited. We should get the retry-after header
        from the response and use it to delay the next request. But this should be implemented
        in python-sdk (pycarlo) library.
        """
        return isinstance(exc, HTTPError) and (
            exc.response and (exc.response.status_code in [408, 429] or exc.response.status_code >= 500)
        )

    def _upload_result_igw(
        self,
        mc_client: Client,
        operation_name: str,
        operation_parameters: Dict,
        payload_value: Dict,
    ):
        body = {
            "airflow_operation": operation_name,
            "airflow_payload": {**operation_parameters, "payload": payload_value},
        }
        try:
            logger.info(f"Sending MC Callback information to Gateway, operation={operation_name}")
            mc_client.make_request(path=_INTEGRATION_GATEWAY_PATH, body=body, timeout_in_seconds=self.mcd_call_timeout,
                                   should_retry=self._should_retry)
        except HTTPError as exc:
            if exc.response.status_code == 401:
                logger.warning(
                    "MCD Gateway authentication failed, please check you created an "
                    "Airflow Integration Key following the docs at "
                    "https://docs.getmontecarlo.com/docs/airflow-incidents-dags-and-tasks"
                )
            logger.exception(f"Failed to upload information to MCD Gateway: {exc}")
        except Exception as exc:
            logger.exception(f"Failed to upload information to MCD Gateway: {exc}")

    def _get_existing_connection_id(self) -> Optional[str]:
        try:
            BaseHook.get_connection(self.mcd_session_conn_id)
            return self.mcd_session_conn_id
        except AirflowNotFoundException:
            pass

        if self.mcd_fallback_conn_id:
            try:
                BaseHook.get_connection(self.mcd_fallback_conn_id)
                logger.warning(f"Using fallback connection id {self.mcd_fallback_conn_id}, please consider "
                               f"switching to {self.mcd_session_conn_id}.")
                return self.mcd_fallback_conn_id
            except AirflowNotFoundException:
                raise AirflowNotFoundException(f"Connection not found, searched for {self.mcd_session_conn_id} "
                                               f"and {self.mcd_fallback_conn_id}")
        else:
            raise AirflowNotFoundException(f"Connection {self.mcd_session_conn_id} not found")

    @staticmethod
    def _get_session(connection_id: str) -> Session:
        return SessionHook(mcd_session_conn_id=connection_id).get_conn()

    @staticmethod
    def _env_to_payload(env: AirflowEnv) -> Dict:
        values = {
            'env_name': env.env_name,
        }
        if env.env_id:
            values['env_id'] = env.env_id
        if env.base_url:
            values['base_url'] = env.base_url
        if env.version:
            values['version'] = env.version

        return values



import pkg_resources


def get_provider_info():
    return {
        "name": "Monte Carlo",
        "description": "`Monte Carlo <https://www.montecarlodata.com/>`__\n",
        "connection-types": [
            {
                "hook-class-name": "airflow_mcd.hooks.SessionHook",
                "connection-type": "mcd",
            },
            {
                "hook-class-name": "airflow_mcd.hooks.GatewaySessionHook",
                "connection-type": "mcd_gateway",
            },
        ],
        "hook-class-names": [
            "airflow_mcd.hooks.SessionHook",
            "airflow_mcd.hooks.GatewaySessionHook",
        ],
        "package-name": "airflow-mcd",
    }


def _check_airflow_version():
    try:
        airflow_dist = pkg_resources.get_distribution("apache-airflow")
    except pkg_resources.DistributionNotFound:
        raise ImportError(
            "apache-airflow is not installed. Please install a compatible version of Airflow "
            "before using airflow_mcd."
        )
    else:
        # If you still need a minimum version, e.g. >=1.10.14:
        if pkg_resources.parse_version(airflow_dist.version) < pkg_resources.parse_version("1.10.14"):
            raise RuntimeError(
                f"Installed apache-airflow=={airflow_dist.version} is too old. "
                "Please upgrade to apache-airflow>=1.10.14."
            )


_check_airflow_version()



from unittest import TestCase
from unittest.mock import patch

from airflow_mcd.operators.base_operator import BaseMcdOperator

SAMPLE_CONN = 'foo'
SAMPLE_CONN_ID = 'mcd_default_session'


class MockSession:
    @staticmethod
    def get_conn():
        return SAMPLE_CONN


class BaseOpTest(TestCase):
    def setUp(self) -> None:
        self._op = BaseMcdOperator(mcd_session_conn_id=SAMPLE_CONN_ID, task_id='test')

    @patch('airflow_mcd.operators.base_operator.SessionHook')
    def test_get_session(self, mock_session):
        mock_session.return_value = MockSession
        self.assertEqual(self._op.get_session(), SAMPLE_CONN)
        mock_session.assert_called_once_with(mcd_session_conn_id=SAMPLE_CONN_ID)



"""
Monte Carlo callbacks module.

You can use the provided helper dictionaries to install MC callbacks, for example
for dags, you can use `dag_callbacks` to install MC callbacks:
::
    from airflow_mcd.callbacks import mcd_callbacks

    dag = DAG(
        'name',
        **mcd_callbacks.dag_callbacks,
    )


For tasks, you can use `task_callbacks` to install MC callbacks in all your tasks using `default_args`:
::
    dag = DAG(
        'name',
        default_args={
            **mcd_callbacks.task_callbacks,
        }
    )

You can also install callbacks on a per-task basis:
::
    task = BashOperator(
        task_id='print_date',
        bash_command='command',
        dag=dag,
        **mcd_callbacks.task_callbacks,
    )

If you have your own callbacks, you can call MC callbacks manually from your code:
::
    def dag_success_alert(context):
        #processing code

        #invoke MC event processing
        mcd_callbacks.mcd_dag_success_callback(context)

Configuring the connection to Monte Carlo: you need to create an HTTP connection in Airflow with name
`mcd_default_session` and the following attributes:

* login: your mcd_id value
* password: your mcd_token value
* or using extra attributes: {"mcd_id": "ID", "mcd_token": "Token"}

If you set the connection name to something that is not the default value ("mcd_default_session")
you need to specify the connection name as a DAG parameter with name `mcd_session_conn_id`, like:
::
    dag = DAG(
        'name',
        params={'mcd_connection_id': 'my_connection_id'}
        **mcd_callbacks.dag_callbacks,
    )
"""

import logging
from typing import Dict

from airflow_mcd.callbacks.utils import AirflowEventsClientUtils

logger = logging.getLogger(__name__)


def mcd_dag_success_callback(context: Dict):
    """
    Callback function that sends the DAG success event to Monte Carlo, it must be configured as
    `on_success_callback` in a DAG object.
    You can also manually invoke this function from your own callback code passing the same context you received.
    :param context: The Airflow DAG success context.
    """
    try:
        AirflowEventsClientUtils.mcd_post_dag_result(context)
    except Exception as ex:
        logger.exception(f'Failed to post dag result to MCD: {ex}')


def mcd_dag_failure_callback(context: Dict):
    """
    Callback function that sends the DAG failure event to Monte Carlo, it must be configured as
    `on_failure_callback` in a DAG object.
    You can also manually invoke this function from your own callback code passing the same context you received.
    :param context: The Airflow DAG failure context.
    """
    try:
        AirflowEventsClientUtils.mcd_post_dag_result(context)
    except Exception as ex:
        logger.exception(f'Failed to post dag failure result to MCD: {ex}')


def mcd_task_success_callback(context: Dict):
    """
    Callback function that sends the Task success event to Monte Carlo, it must be configured as
    `on_success_callback` in a Task object.
    You can also manually invoke this function from your own callback code passing the same context you received.
    :param context: The Airflow Task success context.
    """
    try:
        AirflowEventsClientUtils.mcd_post_task_result(context)
    except Exception as ex:
        logger.exception(f'Failed to post task result to MCD: {ex}')


def mcd_task_execute_callback(context: Dict):
    """
    Callback function that sends the Task execution event to Monte Carlo, it must be configured as
    `on_execute_callback` in a Task object.
    You can also manually invoke this function from your own callback code passing the same context you received.
    :param context: The Airflow Task execution context.
    """
    try:
        AirflowEventsClientUtils.mcd_post_task_result(context)
    except Exception as ex:
        logger.exception(f'Failed to post task result to MCD: {ex}')


def mcd_task_failure_callback(context: Dict):
    """
    Callback function that sends the Task failure event to Monte Carlo, it must be configured as
    `on_failure_callback` in a Task object.
    You can also manually invoke this function from your own callback code passing the same context you received.
    :param context: The Airflow Task failure context.
    """
    try:
        AirflowEventsClientUtils.mcd_post_task_result(context)
    except Exception as ex:
        logger.exception(f'Failed to post task failure result to MCD: {ex}')


def mcd_task_retry_callback(context: Dict):
    """
    Callback function that sends the Task retry event to Monte Carlo, it must be configured as
    `on_retry_callback` in a Task object.
    You can also manually invoke this function from your own callback code passing the same context you received.
    :param context: The Airflow Task retry context.
    """
    try:
        AirflowEventsClientUtils.mcd_post_task_result(context)
    except Exception as ex:
        logger.exception(f'Failed to post task retry result to MCD: {ex}')


def mcd_sla_miss_callback(dag, task_list, blocking_task_list, slas, blocking_tis):
    """
    Callback function that sends the SLA misses event to Monte Carlo, it must be configured as
    `sla_miss_callback` in a DAG object.
    You can also manually invoke this function from your own callback code passing the same arguments
    you received.
    """
    try:
        AirflowEventsClientUtils.mcd_post_sla_misses(dag=dag, sla_misses=slas)
    except Exception as ex:
        logger.exception(f'Failed to post SLA misses result to MCD: {ex}')


task_callbacks = {
    'on_success_callback': mcd_task_success_callback,
    'on_failure_callback': mcd_task_failure_callback,
    'on_retry_callback': mcd_task_retry_callback,
    'on_execute_callback': mcd_task_execute_callback,
}

dag_callbacks = {
    'on_failure_callback': mcd_dag_failure_callback,
    'on_success_callback': mcd_dag_success_callback,
    'sla_miss_callback': mcd_sla_miss_callback,
}



import distutils.text_file
from pathlib import Path
from typing import List, Optional

from setuptools import setup, find_packages


def parse_requirements(file_name: Optional[str] = 'requirements.txt') -> List[str]:
    return distutils.text_file.TextFile(filename=str(Path(__file__).with_name(file_name))).readlines()


def get_long_description():
    with open('README.md', 'r') as fh:
        return fh.read()


setup(
    name='airflow_mcd',
    use_scm_version=True,
    license='Apache Software License (Apache 2.0)',
    description='Monte Carlo\'s Apache Airflow Provider',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    author='Monte Carlo Data, Inc',
    author_email='info@montecarlodata.com',
    url='https://www.montecarlodata.com/',
    packages=find_packages(exclude=['tests*', 'examples*']),
    include_package_data=True,
    install_requires=parse_requirements(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10'
    ],
    python_requires='>=3.7',
    setup_requires=['setuptools', 'setuptools_scm'],
    entry_points=dict(apache_airflow_provider=["provider_info=airflow_mcd:get_provider_info"])
)



from unittest import TestCase
from unittest.mock import patch

from airflow import AirflowException
from airflow.exceptions import AirflowSkipException

from airflow_mcd.operators import SimpleCircuitBreakerOperator

SAMPLE_CONN_ID = 'mcd_default_session'
SAMPLE_RULE_ID = 'foo'
SAMPLE_RULE_NAME = 'rule_name'
SAMPLE_NAMESPACE = 'namespace'
SAMPLE_TIMEOUT_IN_MINUTES = 10
SAMPLE_FAIL_OPEN = False


class CbOpTest(TestCase):
    def setUp(self) -> None:
        self._op = SimpleCircuitBreakerOperator(
            mcd_session_conn_id=SAMPLE_CONN_ID,
            rule_uuid=SAMPLE_RULE_ID,
            timeout_in_minutes=SAMPLE_TIMEOUT_IN_MINUTES,
            fail_open=SAMPLE_FAIL_OPEN,
            task_id='test'
        )

    def test_initialization(self):
        self.assertEqual(self._op.rule_uuid, SAMPLE_RULE_ID)
        self.assertEqual(self._op.timeout_in_minutes, SAMPLE_TIMEOUT_IN_MINUTES)
        self.assertEqual(self._op.fail_open, SAMPLE_FAIL_OPEN)
        self.assertEqual(self._op.mcd_session_conn_id, SAMPLE_CONN_ID)

    @patch.object(SimpleCircuitBreakerOperator, 'get_session')
    @patch('airflow_mcd.operators.circuit_breaker_operators.CircuitBreakerService')
    def test_execute_no_breach(self, cd_mock, get_session_mock):
        cd_mock().trigger_and_poll.return_value = False

        self.assertIsNone(self._op.execute())
        cd_mock().trigger_and_poll.assert_called_once_with(
            rule_uuid=SAMPLE_RULE_ID,
            namespace=None,
            rule_name=None,
            timeout_in_minutes=SAMPLE_TIMEOUT_IN_MINUTES,
            runtime_variables=None
        )
        get_session_mock.assert_called_once_with()

    @patch.object(SimpleCircuitBreakerOperator, 'get_session')
    @patch('airflow_mcd.operators.circuit_breaker_operators.CircuitBreakerService')
    def test_execute_no_breach_with_rule_name_and_namespace(self, cd_mock, get_session_mock):
        cd_mock().trigger_and_poll.return_value = False

        operator = SimpleCircuitBreakerOperator(
            mcd_session_conn_id=SAMPLE_CONN_ID,
            rule_name=SAMPLE_RULE_NAME,
            namespace=SAMPLE_NAMESPACE,
            timeout_in_minutes=SAMPLE_TIMEOUT_IN_MINUTES,
            fail_open=SAMPLE_FAIL_OPEN,
            task_id='test1'
        )

        self.assertIsNone(operator.execute())
        cd_mock().trigger_and_poll.assert_called_once_with(
            rule_uuid=None,
            namespace=SAMPLE_NAMESPACE,
            rule_name=SAMPLE_RULE_NAME,
            timeout_in_minutes=SAMPLE_TIMEOUT_IN_MINUTES,
            runtime_variables=None
        )
        get_session_mock.assert_called_once_with()

    @patch.object(SimpleCircuitBreakerOperator, 'get_session')
    @patch('airflow_mcd.operators.circuit_breaker_operators.CircuitBreakerService')
    def test_execute_no_breach_with_rule_name(self, cd_mock, get_session_mock):
        cd_mock().trigger_and_poll.return_value = False

        operator = SimpleCircuitBreakerOperator(
            mcd_session_conn_id=SAMPLE_CONN_ID,
            rule_name=SAMPLE_RULE_NAME,
            timeout_in_minutes=SAMPLE_TIMEOUT_IN_MINUTES,
            fail_open=SAMPLE_FAIL_OPEN,
            task_id='test2'
        )

        self.assertIsNone(operator.execute())
        cd_mock().trigger_and_poll.assert_called_once_with(
            rule_uuid=None,
            namespace=None,
            rule_name=SAMPLE_RULE_NAME,
            timeout_in_minutes=SAMPLE_TIMEOUT_IN_MINUTES,
            runtime_variables=None
        )
        get_session_mock.assert_called_once_with()

    @patch.object(SimpleCircuitBreakerOperator, 'get_session')
    @patch('airflow_mcd.operators.circuit_breaker_operators.CircuitBreakerService')
    def test_execute_with_breach(self, cd_mock, get_session_mock):
        cd_mock().trigger_and_poll.return_value = True

        with self.assertRaises(AirflowException) as context:
            self._op.execute()
        self.assertEqual(str(context.exception), f'Rule \'{SAMPLE_RULE_ID}\' is in breach!')
        cd_mock().trigger_and_poll.assert_called_once_with(
            rule_uuid=SAMPLE_RULE_ID,
            namespace=None,
            rule_name=None,
            timeout_in_minutes=SAMPLE_TIMEOUT_IN_MINUTES,
            runtime_variables=None
        )
        get_session_mock.assert_called_once_with()

    @patch.object(SimpleCircuitBreakerOperator, 'get_session')
    @patch('airflow_mcd.operators.circuit_breaker_operators.CircuitBreakerService')
    def test_execute_with_breach_and_fail_closed(self, cd_mock, get_session_mock):
        cd_mock().trigger_and_poll.side_effect = ValueError

        with self.assertRaises(AirflowException):
            self._op.execute()
        cd_mock().trigger_and_poll.assert_called_once_with(
            rule_uuid=SAMPLE_RULE_ID,
            namespace=None,
            rule_name=None,
            timeout_in_minutes=SAMPLE_TIMEOUT_IN_MINUTES,
            runtime_variables=None
        )
        get_session_mock.assert_called_once_with()

    @patch.object(SimpleCircuitBreakerOperator, 'get_session')
    @patch('airflow_mcd.operators.circuit_breaker_operators.CircuitBreakerService')
    def test_execute_with_breach_and_fail_open(self, cd_mock, get_session_mock):
        op = SimpleCircuitBreakerOperator(
            mcd_session_conn_id=SAMPLE_CONN_ID,
            rule_uuid=SAMPLE_RULE_ID,
            timeout_in_minutes=SAMPLE_TIMEOUT_IN_MINUTES,
            fail_open=True,
            task_id='test'
        )
        cd_mock().trigger_and_poll.side_effect = ValueError

        with self.assertRaises(AirflowSkipException) as context:
            op.execute()
        cd_mock().trigger_and_poll.assert_called_once_with(
            rule_uuid=SAMPLE_RULE_ID,
            namespace=None,
            rule_name=None,
            timeout_in_minutes=SAMPLE_TIMEOUT_IN_MINUTES,
            runtime_variables=None
        )
        get_session_mock.assert_called_once_with()
        self.assertEqual((str(context.exception)), 'Encountered an error when executing the rule, but failing open.')

    @patch.object(SimpleCircuitBreakerOperator, 'get_session')
    @patch('airflow_mcd.operators.circuit_breaker_operators.CircuitBreakerService')
    def test_execute_with_variables(self, cd_mock, get_session_mock):
        op = SimpleCircuitBreakerOperator(
            mcd_session_conn_id=SAMPLE_CONN_ID,
            rule_uuid=SAMPLE_RULE_ID,
            timeout_in_minutes=SAMPLE_TIMEOUT_IN_MINUTES,
            fail_open=SAMPLE_FAIL_OPEN,
            task_id='test',
            runtime_variables={"var1": "value1"}
        )

        cd_mock().trigger_and_poll.return_value = False

        self.assertIsNone(op.execute())
        cd_mock().trigger_and_poll.assert_called_once_with(
            rule_uuid=SAMPLE_RULE_ID,
            namespace=None,
            rule_name=None,
            timeout_in_minutes=SAMPLE_TIMEOUT_IN_MINUTES,
            runtime_variables={"var1": "value1"}
        )
        get_session_mock.assert_called_once_with()


"""
Example circuit breaker dag using rule UUIDs from the test.snow dev account.

See README-dev.md for details on how to run locally.
"""

from datetime import datetime, timedelta

from airflow import DAG

try:
    from airflow.operators.bash import BashOperator
except ImportError:
    # For airflow versions <= 2.0.0. This module was deprecated in 2.0.0.
    from airflow.operators.bash_operator import BashOperator

from airflow_mcd.operators import SimpleCircuitBreakerOperator

default_args = {
    'owner': 'airflow',
    'depends_on_past': False,
    'email_on_failure': False,
    'email_on_retry': False,
    'retries': 3,
    'retry_delay': timedelta(minutes=5)
}

with DAG(
        'sample-dag',
        default_args=default_args,
        start_date=datetime(2022, 2, 8),
        catchup=False,
        schedule_interval=timedelta(1)
) as dag:
    mcd_connection_id = 'mcd_default_session'

    task1 = BashOperator(
        task_id='example_elt_job_1',
        bash_command='echo I am transforming a very important table!',
    )
    breaker1 = SimpleCircuitBreakerOperator(
        task_id='circuit_breaker_1',
        mcd_session_conn_id=mcd_connection_id,
        rule_uuid='58dcbf8a-056f-4d25-9675-0b7f0d08d218'  # A rule that should always breach
    )
    breaker2 = SimpleCircuitBreakerOperator(
        task_id='circuit_breaker_2',
        mcd_session_conn_id=mcd_connection_id,
        rule_name='monitor_as_code_rule_that_should_never_breach',  # A rule that should never breach
        namespace='monitor_as_code_rules'
    )
    breaker3 = SimpleCircuitBreakerOperator(
        task_id='circuit_breaker_3',
        mcd_session_conn_id=mcd_connection_id,
        rule_name='rule_created_in_the_ui_that_should_never_breach',  # A rule that should never breach
    )
    task2 = BashOperator(
        task_id='example_elt_job_2',
        bash_command='echo I am building a very important dashboard from the table created in task1!',
        trigger_rule='none_failed'
    )

    task1 >> [breaker1, breaker2, breaker3] >> task2
